"""
some modules of transformer
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


def clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])



class Encoder(nn.Module):
    def __init__(self, embed_dim, num_layer, num_heads, ff_dim, dropout):
        super(Encoder, self).__init__()
        self.layers = nn.ModuleList([EncoderLayer(embed_dim, num_heads, ff_dim, dropout)
                                     for _ in range(num_layer)])
        self.norm = LayerNorm(embed_dim)

    def forward(self, h, mask=None):
        for layer in self.layers:
            h = layer(h, mask)
        h = self.norm(h)
        return h


class Decoder(nn.Module):
    def __init__(self, embed_dim, num_layer, num_heads, ff_dim, dropout):
        super(Decoder, self).__init__()
        self.layers = nn.ModuleList([DecoderLayer(embed_dim, num_heads, ff_dim, dropout)
                                     for _ in range(num_layer)])
        self.norm = LayerNorm(embed_dim)

    def forward(self, x, h, self_mask=None, cross_mask=None):
        for i in range(len(self.layers)):
            x = self.layers[i](x, h, self_mask, cross_mask)
        x = self.norm(x)
        return x


class Transformer(nn.Module):

    arch_settings = {
        "nano": (3, 8, 512),
        "small": (6, 8, 512),
        "base": (12, 12, 768),
        "large": (24, 16, 1024),
        "huge": (32, 16, 1280)
    }

    def __init__(self):
        super().__init__()

    def forward(self):
        pass



base_prompt = \
"""Answer the question. Your reply MUST include your answer with tags, e.g., <Answer>A/B/C</Answer>. 

"""

one_shot_cot_prompt = \
"""Please answer the question. Your reply MUST include your answer with tags at the end, e.g., <Answer>A/B/C</Answer>. 

Here is an example: 
\"\"\"
Question: Which word would you find on a dictionary page with the following guide words?
leap - lucky
Context: N/A
Options: (A) lay (B) lord

Answer: 
The guide words at the top of a dictionary page help you to determine what words are listed on that page. The words are listed in alphabetical order. The first guide word is the first word on the page and the second guide word is the last word on the page.

Given the guide words "get" and "goes", any word that fits alphabetically between these two words could be found on this page. Considering the options given:

- (A) group: The word "group" starts with 'gr', which comes after 'go' in alphabetical order, so it would not be found on a page with "get" and "goes" as the guide words.
- (B) gnaw: The word "gnaw" starts with 'gn', which comes after 'ge' but before 'go' in alphabetical order, so it could be found on a page with "get" and "goes" as the guide words.

So, the correct answer is (B) gnaw.

<Answer>B</Answer>
\"\"\"

"""


rationale_line_by_line = \
"""

Let's think step by step:
1. The guide words on a dictionary page hint at the first and last words on that specific page.
2. 'Get' would be the first word on the page.
3. 'Goes' would be the last word on the page.
4. Alphabetically, words on this page would fall between 'get' and 'goes'.
5. 'Group' starts with 'gr-', which comes after 'go-' alphabetically, so it would not be on this page.
6. 'Gnaw' starts with 'gn-', which comes after 'ge-' (from 'get') but before 'go-' (from 'goes') alphabetically, so it would be on this page.

So, the answer is: <Answer>B</Answer>
"""



zeroshot_cot_prompt = \
"""Answer the question after thinking step by step. Your MUST provide your answer with TAGS at the end, e.g., <Answer>A/B/C</Answer>. 
"""


least_to_most_decomposer_prompt = \
"""
- To judge a sentense is true or false, you need to provide a series of sub-questions that lead us to the final answer. 
- You MUST enclose each subquestion with TAGS <subquestion>[your_subquestion]</subquestion>)

There are a few examples: 
Q:
    Judge whether the sentense is true or false: Jane got some weird looks because she wore sunglasses outside at 4 PM.
A:
    To judge the truthfulness of the statement "Jane got some weird looks because she wore sunglasses outside at 4 PM," we would need to know the answers to the following sub-questions:

    <subquestion>Did Jane wear sunglasses outside at 4 PM?</subquestion>
    <subquestion>Did people notice Jane wearing sunglasses?</subquestion>
    <subquestion>Did people give Jane strange looks?</subquestion>
    <subquestion>Is it unusual or socially unacceptable to wear sunglasses outside at 4 PM in the context in which Jane was present?</subquestion>

Q: 
    Judge whether the sentense is true or false: The college coach had been paying his players so he was praised by the media.
A: 
    To judge the truthfulness of the statement "The college coach had been paying his players so he was praised by the media," we would need to know the answers to the following sub-questions:

    <subquestion>If the college coach did pay his players, was this behavior considered acceptable or unethical by the media?</subquestion>
    <subquestion>If the media did praise the college coach, was this praise based solely on his performance as a coach, or did it take his alleged payments to players into account?</subquestion>

Q:
    Judge whether the sentense is true or false: As the weather was very cold he put on his jacket to protect himself. 
A:
    To judge the truthfulness of the statement "As the weather was very cold he put on his jacket to protect himself," we would need to know the answers to the following sub-questions:

    <subquestion>Was the weather very cold?</subquestion>
    <subquestion>Did he have a jacket available to put on?</subquestion>
    <subquestion>Did he put on his jacket as a response to the cold weather?</subquestion>
    <subquestion>Was the purpose of putting on the jacket to protect himself from the cold?</subquestion>

"""


# least_to_most_subq_solver_prompt = \
# """
# You need to answer the question based on the given context.

# An example:
# \"\"\"
# Jane got some weird looks because she wore sunglasses outside at 4 PM.
# Q: Is it confirmed that Jane wore sunglasses outside at 4 PM?
# A: This requires a factual confirmation, either from Jane herself, an eyewitness, or a form of recording like a picture or video.

# Q: Are there any social norms or cultural expectations in Jane's location that might make wearing sunglasses at 4 PM unusual?
# A: This would depend on Jane's geographical location, the local weather, and cultural norms. For example, in sunny regions, wearing sunglasses at 4 PM may be perfectly normal.

# Q: Is there any evidence to support the claim that the looks Jane received were indeed "weird"?
# A: To determine this, we would need detailed accounts from Jane or other eyewitnesses about the reactions she received. 
# \"\"\"

# Now, according to the context, answer the question. 
# """


least_to_most_subq_solver_prompt = \
"""
Answer the questions one by one according to the context. 

There are two examples: 
\"\"\"
Jane got some weird looks because she wore sunglasses outside at 4 PM.
Q: Did Jane wear sunglasses outside at 4 PM?
A: Yes, based on the statement "Jane got some weird looks because she wore sunglasses outside at 4 PM," it can be inferred that Jane did indeed wear sunglasses outside at 4 PM.

Q: Did people notice Jane wearing sunglasses?
A: Yes, based on the statement "Jane got some weird looks", it can be inferred that people noticed Jane wearing sunglasses outside at 4 PM and their reaction was noticeable enough for Jane to feel uncomfortable or self-conscious about it.

Q: Did people give Jane strange looks?
A: Yes, based on the statement "Jane got some weird looks", it can be inferred that people gave Jane strange looks for wearing sunglasses outside at 4 PM.


The college coach had been paying his players so he was praised by the media.
Q: Does the college coach in question actually pay his players?
A: This would involve verifying the information from reliable sources. In the context of collegiate sports, particularly in the United States, this action is typically against NCAA regulations. As of my knowledge cutoff in September 2021, college athletes were not allowed to be paid directly by their coaches or institutions, although recent changes had begun to allow athletes to profit from their name, image, and likeness.

Q: Has this action been made public?
A: We would need to know whether the media is aware of this behavior. If the action is not publicly known, the media would not be in a position to praise or condemn the coach.

Q: Has the media actually praised the coach for this action?
A: This would involve researching media coverage. It's important to note that the media is not a single monolithic entity, so the reactions may vary from one media outlet to another.

Q: If the media has praised the coach, is it specifically for paying the players?
A: The media may praise a coach for various reasons, such as the team's performance, the coach's leadership, etc. We would need to verify that the praise is specifically related to the action of paying the players.

Q: Is the sentense true or false, "The college coach had been paying his players so he was praised by the media"? Your reply MUST include '<Answer>True</Answer>' or '<Answer>False</Answer>' at the end.
A: Based on the overall context and norms within college sports, it's unlikely that a college coach would be praised by the media for paying his players, given that this is typically against the rules of most collegiate sports associations. Therefore, based on the information available to me, I would judge the statement as likely being false. <Answer>False</Answer>
\"\"\"


"""




example0 = \
"""
An example: 
\"\"\"
Question: Which word would you find on a dictionary page with the following guide words?
get - goes
Context: N/A
Options: (A) group (B) gnaw
    
Response: 
- **Explanation of Terms**: 
    - Term 1: Dictionary Page: A dictionary page is a page in a dictionary that contains a list of words and their meanings. The words are usually arranged in alphabetical order.
    - Term 2: Guide Words: Guide words are the words printed at the top of a page in a reference book such as a dictionary or encyclopedia. They help to indicate the first and the last words on that page.

- **Subquestion Decomposition and Answering**: 
    - Subquestion 1: What is the alphabetical position of the guide words and the options?
    - Answer to subquestion 1: The guide words are "get" and "goes". In alphabetical order, "get" comes before "goes". The options are "group" and "gnaw". "Group" comes after "goes" in alphabetical order, and "gnaw" comes before "get".

    - Subquestion 2: Which of the options can be found between the guide words?
    - Answer to subquestion 2: In a dictionary, words are arranged in alphabetical order. Therefore, any word that can be found between the guide words "get" and "goes" would be a word that comes after "get" and before "goes" in alphabetical order. In this case, only "gnaw" fits this criterion.

- **Rationale for Arriving at the Answer**: 
    Based on the alphabetical order, the word "gnaw" comes after "get" and before "goes". Therefore, it is the word that you would find on a dictionary page with the guide words "get - goes". The word "group" comes after "goes", so it would not be on this page.

- **Provide Your Answer WITH TAGS**: 
    <Answer>B</Answer>
\"\"\"
"""


example1 = \
"""
An example: 
\"\"\"
Question: Which word would you find on a dictionary page with the following guide words?
get - goes
Context: N/A
Options: (A) group (B) gnaw
    
Response: 
- **Explanation of Terms**: 
    - Term 1: Dictionary Page: A dictionary page is a page in a dictionary that contains a list of words and their meanings. The words are usually arranged in alphabetical order.
    - Term 2: Guide Words: Guide words are the words printed at the top of a page in a reference book such as a dictionary or encyclopedia. They help to indicate the first and the last words on that page.

- **Subquestion Decomposition and Answering**: 
    - Subquestion 1: What is the alphabetical position of the guide words and the options?
    - Answer to subquestion 1: The guide words are "get" and "goes". In alphabetical order, "get" comes before "goes". The options are "group" and "gnaw". "Group" comes after "goes" in alphabetical order, and "gnaw" comes before "get".

    - Subquestion 2: Which of the options can be found between the guide words?
    - Answer to subquestion 2: In a dictionary, words are arranged in alphabetical order. Therefore, any word that can be found between the guide words "get" and "goes" would be a word that comes after "get" and before "goes" in alphabetical order. In this case, only "gnaw" fits this criterion.

- **Rationale for Arriving at the Answer**: 
    Based on the alphabetical order, the word "gnaw" comes after "get" and before "goes". Therefore, it is the word that you would find on a dictionary page with the guide words "get - goes". The word "group" comes after "goes", so it would not be on this page.

- **Provide Your Answer WITH TAGS**: 
    <Answer>B</Answer>
\"\"\"
"""

# example0 = ''
# example1 = ''

# our_reasoner1_prompt = \
# f"""
# Follow the outline to solve the question: 
# - **List necessary information to solve the question**: 
# - **Explanation of Terms**: 
# - **Rationale**: 
# - **Provide your answer WITH TAGS**: 
# (You MUST choose ONE option, with the capital letter included in TAGS, e.g., <Answer>A/B/C</Answer>. If the answer cannot be determined, make an educated guess. )


# {example0}

# Now, solve the question below by following the outline. Remember, You MUST give a certain answer, with the capital letter included in TAGS, e.g., <Answer>A/B/C</Answer>. If the answer cannot be determined, make an educated guess. 

# """



# our_reasoner2_prompt = \
# f"""
# Follow the outline to solve the question: 
# - **List necessary information to solve the question**: 
# - **Explanation of Terms**: 
# - **Rationale**: 
# - **Provide your answer WITH TAGS**: 
# (You MUST choose ONE option, with the capital letter included in TAGS, e.g., <Answer>A/B/C</Answer>. If the answer cannot be determined, make an educated guess. )


# {example1}

# Now, solve the question below by following the outline. Remember, You MUST give a certain answer, with the capital letter included in TAGS, e.g., <Answer>A/B/C</Answer>. If the answer cannot be determined, make an educated guess. 

# """




our_reasoner1_prompt = \
f"""
Follow the outline below to solve the question: 
    - **Explanation of Terms**: 
    (You need to explain each term used in the question to remove ambiguity. )
        - Term 1: ...

        - Term 2: ...

    - **Subquestion Decomposition and Answering**: 
    (You need to decompose the question into several subquestions connected logically to arrive at the final answer. For each subquestion, provide your answer below it. )
        - Subquestion 1: ...
        - Answer to subquestion 1: ...
        
        - Subquestion 2: ...
        - Answer to subquestion 2: ...

    - **Rationale for Arriving at the Answer**: 
    (You need to reason step by step for the correct answer based on the previous information. )

    - **Provide Your Answer WITH TAGS**: 
    (You MUST choose ONE option, with the capital letter included in TAGS, e.g., <Answer>A/B/C</Answer>. If the answer cannot be determined, make an educated guess. )


{example0}

Now, solve the question below by following the outline. Remember, You MUST give a certain answer, with the capital letter included in TAGS, e.g., <Answer>A/B/C</Answer>. If the answer cannot be determined, make an educated guess. 
"""


"""
  - **Subquestion Decomposition and Answering**: 
  (You need to decompose the question into several subquestions connected logically. )
    - Subquestion 1: ...
    - Subquestion 2: ...

  - **Subquestion Decomposition and Answering**: 
  (You need to decompose the question into several subquestions connected logically to arrive at the final answer. For each subquestion, provide your answer below it. )
    - Subquestion 1: ...
      - Answer to subquestion 1: ...
    - Subquestion 2: ...
      - Answer to subquestion 2: ...
"""

our_reasoner2_prompt = \
f"""
Follow the outline below to solve the question: 
    - **Explanation of Terms**: 
    (You need to explain each term used in the question to remove ambiguity. )
        - Term 1: ...

        - Term 2: ...

    - **Subquestion Decomposition and Answering**: 
    (You need to decompose the question into several subquestions connected logically to arrive at the final answer. For each subquestion, provide your answer below it. )
        - Subquestion 1: ...
        - Answer to subquestion 1: ...
        
        - Subquestion 2: ...
        - Answer to subquestion 2: ...

    - **Rationale for Arriving at the Answer**: 
    (You need to reason step by step for the correct answer based on the previous information. )

    - **Provide Your Answer WITH TAGS**: 
    (You MUST choose ONE option, with the capital letter included in TAGS, e.g., <Answer>A/B/C</Answer>. If the answer cannot be determined, make an educated guess. )


{example0}

Now, solve the question below by following the outline. Remember, You MUST give a certain answer, with the capital letter included in TAGS, e.g., <Answer>A/B/C</Answer>. If the answer cannot be determined, make an educated guess. 
"""


reviewer_example = \
"""
Here is an example: 
For the question: 
\"\"\"
Question: Which word would you find on a dictionary page with the following guide words?
get - goes
Context: N/A
Options: (A) group (B) gnaw
\"\"\"


and a solution: 
\"\"\"
- **Explanation of Terms**:
    - Dictionary: a book or electronic resource that lists the words of a language in alphabetical order and gives their meaning, or that gives the equivalent words in a different language.
    
    - Guide words: the words printed at the top of a dictionary page indicating the first and last entries on that page.

- **Subquestion Decomposition and Answering**:

    - Subquestion 1: What are guide words in a dictionary?
    - Answer to subquestion 1: Guide words are the words printed at the top of a dictionary page indicating the first and last entries on that page.

    - Subquestion 2: What are the guide words for the dictionary page that contains the word "get" and "goes"?
    - Answer to subquestion 2: The guide words for the dictionary page that contains the word "get" and "goes" are "get" (the first word) and "goes" (the last word).

    - Subquestion 3: Which word would you find on a dictionary page with the guide words "get" and "goes"?
    - Answer to subquestion 3: Based on the guide words "get" and "goes," the word that would appear on the dictionary page is between those two words in alphabetical order.

    - Subquestion 4: Which word is between "get" and "goes" in alphabetical order?
    - Answer to subquestion 4: The word between "get" and "goes" in alphabetical order is "group."

    - Subquestion 5: Which word would you find on a dictionary page with the guide words "get" and "goes"?
    - Answer to subquestion 5: The word that you would find on a dictionary page with the guide words "get" and "goes" is "group."

- **Rationale for Arriving at the Answer**:
    
    - The question asks which word would you find on a dictionary page with the guide words "get" and "goes." To answer this question, we need to understand what guide words are and what the guide words for the dictionary page that contains the word "get" and "goes" are. The guide words for this page are "get" and "goes," which means that the word we are looking for is between these two words in alphabetical order. The only option that is between "get" and "goes" in alphabetical order is "group," so the answer is (A) group.

- **Provide Your Answer WITH TAGS**: 
    
    - <Answer>A</Answer>
\"\"\"

with the option that you need to look out for is (B). 


Your response is like:
\"\"\"
**Review Statements One by One**

    - Evaluation on Explanation of Terms: 
    The terms "Dictionary" and "Guide words" are correctly defined, aiding a user's understanding of the context and the question.

    - Evaluation on Subquestion Decomposition and Answering: 
    The decomposition of the question into subquestions is done systematically and answered appropriately. However, subquestion 5 is redundant as it is identical to subquestion 3.

    - Evaluation on the Reasoning process: 
    The reasoning process is not logically sound in this instance. The solution incorrectly concludes that the word "group" comes between "get" and "goes" in alphabetical order. The correct word from the options given is "gnaw," which does come in between "get" and "goes".

    - Evaluation on the Answer: 
    The answer is incorrect. The word "group" does not come between "get" and "goes" in alphabetical order. The correct word from the options given is "gnaw," which does come in between "get" and "goes".

**Reconsider the Question Step by Step and Consider Counterfactuals**
    - Your counterfactual reasoning about the option that you need to look out for:

        - The option that needs to be looked out for is: (B) gnaw.
        - What if we apply this option: If we consider "gnaw" as the answer, it fits neatly between "get" and "goes" in alphabetical order, unlike the solution's proposed "group" which actually comes after "goes."
        - Will there be a contradiction: No, there is no contradiction in this case. In fact, it corrects the initial mistake of overlooking "gnaw" and incorrectly proposing "group" as the word between "get" and "goes."

    - Your step-by-step reasoning to arrive at the most likely answer:

        - First, we are asked to find the word that falls between "get" and "goes" alphabetically.
        - We then compare the options available to us.
        - Upon comparison, we find that "group" comes after "goes," which is not what we are looking for.
        - However, "gnaw" falls between "get" and "goes" alphabetically.
        - Therefore, we can conclude that the most likely answer is (B) gnaw.

**Provide Your Answer With TAGS**

    - Answer: <Answer>B</Answer>
\"\"\"
"""

# reviewer_example = ''

# our_reviewer_prompt = \
# f"""
# You are an outstanding reviewer. You are responsible for reviewing a possible solution following the outline: 
# ### Review Statements One by One
# (Especially, take care of Factualness and Reasoning Errors in the Solution)

# ### Reconsider the Question Step by Step and Options Counterfactually
# (Reason step by step to arrive at the most likely answer. Meanwhile, consider what if we apply other options. )

# ### Provide Your Answer With TAGS
# (At the end, you MUST choose ONE option as your final answer and mark it with TAGS, e.g., <Answer>A/B/C/D/E</Answer>. Make an educated guess if the answer CANNOT BE DETERMINED. )

# {reviewer_example}

# """

our_reviewer_prompt = \
f"""
You are an objective and fair reviewer. You are responsible for evaluating a possible solution following the outline: 
    - **Review Statements One by One**
        (Examine the following aspects of the solution. Especially, watch out for factualness errors and inference errors in the solution)
        - Evaluation of Explanation of Terms: 
        - Evaluation of Subquestion Decomposition and Answering: 
        - Evaluation of the Reasoning process: 
        - Evaluation of the Answer: 

    - **Reconsider the Question Step by Step and Consider Counterfactuals**
        - Your counterfactual reasoning about the answer that you need to look out for:   
            - The option that needs to be looked out for is: 
            - What if we apply this option: 
            - Will there be a contradiction: 
        - Your step-by-step reasoning to arrive at the most likely answer: 

    - **Provide Your Answer With TAGS**
      (You MUST enclose your answer with TAGS, e.g., <Answer>A/B/C</Answer>. Make an educated guess if the answer CANNOT BE DETERMINED. )

      
{reviewer_example}

"""





term_eval_prompt = \
"""
You are an objective and fair reviewer. You are responsible for evaluating the definition of some given terms following the template below: 
    - **Evaluation on Explanation of Terms**
        - Evaluation on Term 1: 
        - Evaluation on Term 2: 
        ...
    - **Conclusion**
        - <Evaluation>Correct/Partially Correct/Incorrect</Evaluation>
        - (If your evaluate the explanation as Partially Correct/Incorrect, you need to explain your evaluation here. )

"""


import torch
from torch import nn
import numpy as np
from modules.prompt.caco_cot import *
import requests
import re
import random
from modules.prompt.scienceqa_base_prompt import *


class CoT(object):
    def __init__(self, cfgs):
        self.cfgs = cfgs

    def get_single_run_claude(self, cot_prompt, prompt): 
        """
        anthropic = Anthropic(api_key=args.key)
        response = anthropic.completions.create(
            model="claude-1",
            temperature=args.temperature,
            top_p=args.top_p,
            max_tokens_to_sample=3000,
            prompt=f"{HUMAN_PROMPT} {prompt+cot_prompt} {AI_PROMPT}",
        )

        return response.completion
        """
        pass

    def get_single_run_gpt(self, cot_prompt, prompt): 
        headers = {
                'Authorization': f'Bearer ' + self.cfgs["api_key"], 
                'Content-Type': 'application/json'
                }

        response = requests.post(self.cfgs["api_url"], 
                                headers=headers, 
                                json={
                                    "model": self.cfgs["llm"],
                                    "messages": [
                                        {'role': 'user', 'content': prompt}, 
                                        ],
                                    "temperature": self.cfgs["temperature"],
                                    "top_p": self.cfgs["top_p"],
                                    }).json()
        
        return response['choices'][0]['message']['content']

    def filter_output(self, output, choice):

        tags = re.findall(r'<Answer>(.*?)</Answer>', output)
        if len(tags) == 0:
            answer = output
        else:
            answer = tags[-1]
        
        num_choice = len(choice)
        # print(choice)
        if len(answer) == 1 and answer in ['A', 'B', 'C', 'D', 'E'][:num_choice]:
            return answer[0]
        
        elif answer in choice:
            return ['A', 'B', 'C', 'D', 'E'][choice.index(answer)]
        
        elif 'cannot' in answer.lower():
            return ['A', 'B', 'C', 'D', 'E'][random.choice(range(num_choice))]
        
        else: 
            answer = [a for a in ['A', 'B', 'C', 'D', 'E'][:num_choice] if a in answer]
            # print(answer)
            if len(answer) == 1: 
                return answer[-1]
            elif len(answer) > 1:
                return answer[random.choice(range(len(answer)))]
            else:
                return ['A', 'B', 'C', 'D', 'E'][random.choice(range(num_choice))]

    def build_prompt_sciqa(self, problems, qid):
        question = get_question_text(problems[qid])
        context = get_context_text(problems[qid], False)
        choice = get_choice_text(problems[qid], self.cfgs["options"])
        answer = get_answer(problems[qid], self.cfgs["options"])
        lecture = get_lecture_text(problems[qid])
        solution = get_solution_text(problems[qid])
        test_example = create_one_example(self.cfgs["prompt_format"],
                                        question,
                                        context,
                                        choice,
                                        answer,
                                        lecture,
                                        solution,
                                        test_example=True)
        # print(test_example)
        # exit(0)
        method = self.cfgs["method"]
        if method == 'base':
            prompt = base_prompt + test_example
        elif method in ['zeroshot-cot', 'zs-self-consistent', 'zs-complexity']:
            prompt = zeroshot_cot_prompt + test_example
        elif method in ['oneshot-cot', 'os-self-consistent', 'os-complexity']:
            prompt = one_shot_cot_prompt + test_example # cot_prompt + 
        elif method == 'caco_cot':
            prompt = f"\"\"\"\n{test_example}\n\"\"\""
        elif method == 'least-to-most':
            prompt = least_to_most_decomposer_prompt + \
            f"""
            Now, provide the sub-questions for the question: 
            Q: 
                Judge whether the sentense is true or false: {test_example}"""

        else:
            raise NotImplementedError('Method not implemented')
        return prompt

    def __call__(self, qid, prompt, choice):
        get_single_run = self.get_single_run_gpt
        prompt = ''
        cot_prompt = ''
        if 'self-consistent' in self.cfgs["method"] or 'complexity' in self.cfgs["method"]:
            answers = []
            outputs = ''
            n_chains = []
            for i in range(10): 
                cot_prompt = 'Let\'s think step by step. '
                response = get_single_run(cot_prompt, prompt)
                output = response

                n_chain = output.count('\n') - output.count('\n\n') + 1
                n_chains.append(n_chain) # \n or \n\n + 1 as number of chains
                outputs += '\n' + '-'*50 + f'\n\nRound {i+1} #Chain {n_chain} \n Output: {output}'
                answer = self.filter_output(output, choice)
                answers.append(answer)

                # print(answer)

                
            if 'complexity' in self.cfgs["method"]: # filter out those with a shorter reasoning chain
                longchain_indices = np.argsort(n_chains)[-int(0.6*len(n_chains)):] # indices of those with more chains
                answers = np.array(answers)[longchain_indices]

            answer = max(set(answers), key=list(answers).count)
            output = outputs
            # print(answer, output)

        elif self.cfgs["method"] in ['zeroshot-cot']:
            output = get_single_run('Let\'s think step by step. ', prompt)
            answer = self.filter_output(output, choice)

        elif self.cfgs["method"] in ['base', 'oneshot-cot', 'zeroshot-cot']:
            output = get_single_run('', prompt)
            answer = self.filter_output(output, choice)

        elif self.cfgs["method"] == 'least-to-most':
            # PROBLEM DECOMPOSITION
            response = get_single_run('', prompt)
            output = response
            
            subquestion_list = re.findall(r'<subquestion>(.*?)</subquestion>', output)

            original_question = prompt.split('Judge whether the sentense is true or false: ')[-1]
            messages = f"{least_to_most_subq_solver_prompt}Now, the context is: \n{original_question}"
            # PROBLEM SOLVING {original_question } {AI_PROMPT}
            for subq in subquestion_list: 
                messages += '\nQ: ' + subq
                
                ans = get_single_run('', messages)
                if 'A:' not in ans:
                    messages += '\nA: ' + ans + '\n\n'
                else:
                    messages += '\nA:' + ans.split('A:')[1] + '\n\n'
            
            messages += f"\nQ: Is the sentense true or false, \"{original_question}\"? Your reply MUST include '<Answer>True</Answer>' or '<Answer>False</Answer>' at the end."
            ans = get_single_run('', messages)
            messages += '\nA: ' + ans + '\n\n'
            answer = self.filter_output(ans, choice)
            output = messages.split(f'Now, the context is: \n{original_question}')[1]
            
        return qid, answer, output



import torch
import numpy as np
from torchvision import transforms
from torch.utils.data import DataLoader
import os
import json
from PIL import Image
from torch.utils.data import Dataset



class BaseDataset(Dataset):
    def __init__(self, cfgs, tokenizer, split, transform=None):
        self.image_dir = cfgs["image_dir"]
        self.ann_path = cfgs["ann_path"]
        self.max_seq_length = cfgs["max_seq_length"]
        self.split = split
        self.tokenizer = tokenizer
        self.transform = transform
        self.ann = json.loads(open(self.ann_path, 'r').read())

        self.examples = self.ann[self.split]
        if cfgs["dataset_name"] == 'ffa_ir': self.dict2list4ffair()
        for i in range(len(self.examples)):
            self.examples[i]['ids'] = tokenizer(self.examples[i]['report'])[:self.max_seq_length]
            self.examples[i]['mask'] = [1] * len(self.examples[i]['ids'])

    def __len__(self):
        return len(self.examples)

    def dict2list4ffair(self):
        examples_list = []
        for k, v in self.examples.items():
            v['id'] = k
            v['image_path'] = v.pop('Image_path')
            v['report'] = v.pop('En_Report')
            examples_list.append(v)
        self.examples = examples_list


class IuxrayMultiImageDataset(BaseDataset):
    def __getitem__(self, idx):
        example = self.examples[idx]
        image_id = example['id']
        image_path = example['image_path']
        image_1 = Image.open(os.path.join(self.image_dir, image_path[0])).convert('RGB')
        image_2 = Image.open(os.path.join(self.image_dir, image_path[1])).convert('RGB')
        if self.transform is not None:
            image_1 = self.transform(image_1)
            image_2 = self.transform(image_2)
        image = torch.stack((image_1, image_2), 0)
        report_ids = example['ids']
        report_masks = example['mask']
        seq_length = len(report_ids)
        sample = (image_id, image, report_ids, report_masks, seq_length)
        return sample


class MimiccxrSingleImageDataset(BaseDataset):
    def __getitem__(self, idx):
        example = self.examples[idx]
        image_id = example['id']
        image_path = example['image_path']
        image = Image.open(os.path.join(self.image_dir, image_path[0])).convert('RGB')
        if self.transform is not None:
            image = self.transform(image)
        report_ids = example['ids']
        report_masks = example['mask']
        seq_length = len(report_ids)
        sample = (image_id, image, report_ids, report_masks, seq_length)
        return sample


class FFAIRDataset(BaseDataset):
    def __getitem__(self, idx):
        example = self.examples[idx]
        image_id = example['id']
        image_path = example['image_path']
        # image_path = eval(self.image_path[case_id])
        images = []
        count_img = len(image_path)
        for ind in range(count_img):
            # print("image_path[ind]",image_path[ind])
            image = Image.open(os.path.join(self.image_dir, image_path[ind])).convert('RGB')
            # im_shape = image.size  # (512, 512)

            # im_size_min = np.min(im_shape[0:2])
            # im_shapes.append(im_shape)
            # im_scales.append(self.height / im_size_min)
            if self.transform is not None:
                images.append(self.transform(image))

        images = torch.stack(images, 0)
        # im_info = [self.height, self.width, im_scales[0]]
        # im_info = torch.Tensor(im_info)

        reports_ids = example['ids']
        # print(reports_ids)
        reports_masks = example['mask']
        max_seq_length = len(reports_ids)

        """
        if count_img >= 96:
            # random select
            _index = np.arange(0, count_img)
            np.random.shuffle(_index)
            _index = _index[:96]
            images = images[_index]
        else:
            # all + random select
            _index = np.random.randint(0, count_img, 96 - count_img)
            images = torch.cat([images, images[_index]], 0)
        """
        if self.split == 'train':
            _index = torch.randint(0, len(image_path), [8])
            images = images[_index]
        else:
            if len(image_path) > 96:
                image_id = image_id[:96]
                images = images[:96]

        sample = (image_id, images, reports_ids, reports_masks, max_seq_length)
        return sample


class MixSingleImageDataset(Dataset):
    def __init__(self, cfgs, tokenizer, split, transform=None):
        self.image_dir = {'iu_xray': cfgs['image_dir']['iu_xray'],
                          'mimic_cxr': cfgs['image_dir']['mimic_cxr']}
        self.ann_path = {'iu_xray': cfgs['ann_path']['iu_xray'],
                          'mimic_cxr': cfgs['ann_path']['mimic_cxr']}
        self.max_seq_length = cfgs["max_seq_length"]
        self.split = split
        self.tokenizer = tokenizer  # vocab + <unk>
        self.transform = transform
        self.ann = {'iu_xray': json.loads(open(self.ann_path['iu_xray'], 'r').read()),
                    'mimic_cxr': json.loads(open(self.ann_path['mimic_cxr'], 'r').read())}

        self.examples = self.ann['iu_xray'][self.split]
        length = len(self.examples)
        for i in range(length):
            self.examples[i]['ids'] = tokenizer(self.examples[i]['report'], dataset='iu_xray')[:self.max_seq_length]
            self.examples[i]['mask'] = [1] * len(self.examples[i]['ids'])

        self.examples += self.ann['mimic_cxr'][self.split]
        for i in range(length, len(self.examples)):
            self.examples[i]['ids'] = tokenizer(self.examples[i]['report'], dataset='mimic_cxr')[:self.max_seq_length]
            self.examples[i]['mask'] = [1] * len(self.examples[i]['ids'])

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, idx):
        example = self.examples[idx]
        image_id = example['id']
        image_path = example['image_path']
        if len(image_path) > 1:
            if torch.rand(1) > 0.5:
                image = Image.open(os.path.join(self.image_dir['iu_xray'], image_path[0])).convert('RGB')
            else:
                image = Image.open(os.path.join(self.image_dir['iu_xray'], image_path[1])).convert('RGB')
            if self.transform is not None:
                image = self.transform(image)
        else:
            image = Image.open(os.path.join(self.image_dir['mimic_cxr'], image_path[0])).convert('RGB')
            if self.transform is not None:
                image = self.transform(image)
        report_ids = example['ids']
        report_masks = example['mask']
        seq_length = len(report_ids)
        sample = (image_id, image, report_ids, report_masks, seq_length)
        return sample




class MRGDataLoader(DataLoader):
    def __init__(self, cfgs, tokenizer, split, shuffle):
        self.cfgs = cfgs
        self.dataset_name = cfgs["dataset_name"]
        self.batch_size = cfgs["batch_size"]
        self.shuffle = shuffle
        self.num_workers = cfgs["num_workers"]
        self.tokenizer = tokenizer
        self.split = split

        # only ffa ir val and test set use 1 batch size
        if self.dataset_name == 'ffa_ir' and self.split != 'train':
            self.batch_size = 1

        if split == 'train':
            self.transform = transforms.Compose([
                transforms.Resize(256),
                transforms.RandomCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize((0.485, 0.456, 0.406),
                                     (0.229, 0.224, 0.225))])
        else:
            self.transform = transforms.Compose([
                transforms.Resize((224, 224)),
                transforms.ToTensor(),
                transforms.Normalize((0.485, 0.456, 0.406),
                                     (0.229, 0.224, 0.225))])

        if self.dataset_name == 'iu_xray':
            self.dataset = IuxrayMultiImageDataset(self.cfgs, self.tokenizer, self.split, transform=self.transform)
        elif self.dataset_name == 'mimic_cxr':
            self.dataset = MimiccxrSingleImageDataset(self.cfgs, self.tokenizer, self.split, transform=self.transform)
        elif self.dataset_name == 'ffa_ir':
            self.dataset = FFAIRDataset(self.cfgs, self.tokenizer, self.split, transform=self.transform)
        elif self.dataset_name == 'mix':
            self.dataset = MixSingleImageDataset(self.cfgs, self.tokenizer, self.split, transform=self.transform)
        else:
            raise ValueError

        self.init_kwargs = {
            'dataset': self.dataset,
            'batch_size': self.batch_size,
            'shuffle': self.shuffle,
            'collate_fn': self.collate_fn,
            'num_workers': self.num_workers
        }
        super().__init__(**self.init_kwargs)

    @staticmethod
    def collate_fn(data):
        images_id, images, reports_ids, reports_masks, seq_lengths = zip(*data)
        images = torch.stack(images, 0)
        max_seq_length = max(seq_lengths)

        targets = np.zeros((len(reports_ids), max_seq_length), dtype=int)
        targets_masks = np.zeros((len(reports_ids), max_seq_length), dtype=int)

        for i, report_ids in enumerate(reports_ids):
            targets[i, :len(report_ids)] = report_ids

        for i, report_masks in enumerate(reports_masks):
            targets_masks[i, :len(report_masks)] = report_masks

        return images_id, images, torch.LongTensor(targets), torch.FloatTensor(targets_masks)




import os
from abc import abstractmethod
import time

import numpy as np
import torch
import pandas as pd
from utils.monitor import Monitor


class BasePipeline(object):
    def __init__(self, model, cfgs):
        self.cfgs = cfgs

        # setup GPU device if available, move model into configured device
        # TODO multi-gpu
        self.model = model.cuda() if torch.cuda.is_available() else model
        
        self.optimizer = None
        self.criterion = None
        self.metric_caculator = None

        # monitor
        self.monitor = Monitor(cfgs)

        self.checkpoint_dir = self.cfgs["result_dir"]
        if not os.path.exists(self.checkpoint_dir):
            os.makedirs(self.checkpoint_dir)

        self.start_epoch = 1
        if self.cfgs["resume"] != "":
            self._resume_checkpoint(self.cfgs["resume"])


    @abstractmethod
    def _train_epoch(self, epoch):
        raise NotImplementedError

    def train(self):
        for epoch in range(self.start_epoch, self.cfgs['epochs'] + 1):
            self._train_epoch(epoch)

            # save logged informations into log dict
            self.monitor.record_best()

            # print logged informations to the screen
            for key, value in self.monitor.name2val.items():
                print('\t{:15s}: {}'.format(str(key), value))

            # evaluate model performance according to configured metric, save best checkpoint as model_best
            is_best, early_stop = self.monitor.check_best()
            if early_stop:
                break

            if epoch % self.cfgs['save_period'] == 0:
                self._save_checkpoint(epoch, save_best=is_best)
            
            self.monitor.dump(epoch)

        self.monitor.print_best_and_save_to_file()
    
    def inference(self, loader):
        pass

    def _save_checkpoint(self, epoch, save_best=False):
        state = {
            'epoch': epoch,
            'state_dict': self.model.state_dict(),
            'optimizer': self.optimizer.state_dict() if self.optimizer is not None else self.optimizer,
            'monitor_best': self.monitor.monitor_best,
            'seed': self.cfgs['seed']
        }
        filename = os.path.join(self.checkpoint_dir, 'current_checkpoint.pth')
        torch.save(state, filename)
        print("Saving checkpoint: {} ...".format(filename))
        if save_best:
            best_path = os.path.join(self.checkpoint_dir, 'model_best.pth')
            torch.save(state, best_path)
            print("*************** Saving current best: model_best.pth ... ***************")

    def _resume_checkpoint(self, resume_path):
        resume_path = str(resume_path)
        print("Loading checkpoint: {} ...".format(resume_path))
        checkpoint = torch.load(resume_path)
        self.start_epoch = checkpoint['epoch'] + 1
        self.monitor.monitor_best = checkpoint['monitor_best']
        self.model.load_state_dict(checkpoint['state_dict'])
        
        assert self.optimizer is not None
        self.optimizer.load_state_dict(checkpoint['optimizer'])

        print("Checkpoint loaded. Resume training from epoch {}".format(self.start_epoch))



import os
import json
import random



def load_scienceqa_data(cfgs):
    problems = json.load(open(os.path.join(cfgs["data_dir"], 'problems.json')))
    pid_splits = json.load(open(os.path.join(cfgs["data_dir"], 'pid_splits.json')))

    qids = pid_splits['%s' % (cfgs["test_split"])]
    qids = qids[:cfgs["test_number"]] if cfgs["test_number"] > 0 else qids
    print(f"number of test problems: {len(qids)}\n")

    if cfgs["txt_only"]:
        txt_qids = []
        for qid in qids:
            if problems[qid]['image'] == None:
                txt_qids.append(qid) 
        print('Number of text-only questions: ', len(txt_qids))

    # pick up shot examples from the training set
    shot_qids = cfgs["shot_qids"]
    train_qids = pid_splits['train']
    if shot_qids == None:
        assert cfgs["shot_number"] >= 0 and cfgs["shot_number"] <= 32
        shot_qids = random.sample(train_qids, cfgs["shot_number"])  # random sample
    else:
        shot_qids = [str(qid) for qid in shot_qids]

    print("training question ids for prompting: ", shot_qids, "\n")

    return problems, txt_qids, shot_qids



"""
optimizers and lr scheduler
"""
import torch
import math


def build_optimizer(cfgs, model):
    task = cfgs["task"]
    optim = cfgs["optim"]
    weight_decay = cfgs["weight_decay"]
    lr_en = cfgs["lr_en"]
    lr_de = cfgs["lr_de"]
    amsgrad = cfgs["amsgrad"]
    if task == 'mrg':
        en_params = list(map(id, model.vis_embed.parameters()))
        de_params = filter(lambda x: id(x) not in en_params, model.parameters())
        optimizer = getattr(torch.optim, optim)(
            [{'params': model.vis_embed.parameters(), 'lr': lr_en, "lr_scale": 1.},
             {'params': de_params, 'lr': lr_de}],
            weight_decay=weight_decay,
            amsgrad=amsgrad,
        )
    else:
        optimizer = getattr(torch.optim, optim)(
            [{'params': model.parameters(), 'lr': lr_de}],
            weight_decay=weight_decay,
            amsgrad=amsgrad,
        )
    return optimizer


def build_lr_scheduler(cfgs, optimizer, l):
    lr_scheduler_fn = cfgs["lr_scheduler"]
    step_size = cfgs["step_size"]
    epochs = cfgs["epochs"]
    if lr_scheduler_fn == 'warmup_steplr':
        lr_scheduler = WarmupAndSteplr(optimizer, step_size * l, epochs * l)
    elif lr_scheduler_fn == 'warmup':
        lr_scheduler = get_linear_schedule_with_warmup(optimizer, step_size * l, epochs * l)
    else:
        lr_scheduler = getattr(torch.optim.lr_scheduler, cfgs["lr_scheduler"])(optimizer, step_size, cfgs["gamma"])
    print(f"Build {lr_scheduler_fn} for {cfgs['optim']} in {cfgs['task']}")
    return lr_scheduler


class WarmupAndSteplr(object):
    def __init__(self, optimizer, num_warmup_steps, num_training_steps, last_epoch=-1):
        self.end_lr = 0.001
        self.current_step = 0
        self.optim = optimizer
        self.num_warmup_steps = num_warmup_steps
        self.num_training_steps = num_training_steps
        self.lr_ve = self.optim.param_groups[0]['lr']
        self.lr_ed = self.optim.param_groups[-1]['lr']

    def step(self):
        self.current_step += 1
        if self.current_step < self.num_warmup_steps:
            lr_1 = float(self.current_step) / float(max(1, self.num_warmup_steps))
        else:
            # cosine annealing decay
            progress = float(self.current_step - self.num_warmup_steps) / float(
                max(1, self.num_training_steps - self.num_warmup_steps))
            cosine_lr = max(0.0, 0.5 * (1. + math.cos(math.pi * progress)))
            # lr = max(0.0, cosine_lr * (base_lr - end_lr) + end_lr)
            lr_1 = max(0.0, cosine_lr * (1 - self.end_lr) + self.end_lr)

        # if self.current_step < self.num_warmup_steps:
        #     lr_2 = lr_1
        if self.current_step == self.num_training_steps // 2:
            lr_2 = .1
        else:
            lr_2 = 1.

        for param_group in self.optim.param_groups:
            if "lr_scale" in param_group:
                param_group["lr"] = lr_1 * param_group["lr_scale"] * self.lr_ve
            else:
                param_group["lr"] = lr_2 * self.lr_ed
        return lr_1 * self.lr_ve, lr_2 * self.lr_ed


def get_linear_schedule_with_warmup(optimizer, num_warmup_steps, num_training_steps, last_epoch=-1):
    """
    when current_step < num_warmup_steps,
    new_lr =current_step/num_warmup_steps * base_lr
    when current_step >= num_warmup_steps,
    new_lr =(num_training_steps - current_step) / (num_training_steps -num_warmup_steps) * base_lr

    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The total number of training steps.
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.

    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_line(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
        )

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
        )

    def lr_cosine(current_step: int):
        # linear warmup
        end_lr = 0.001
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        # cosine annealing decay
        progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
        cosine_lr = max(0.0, 0.5 * (1. + math.cos(math.pi * progress)))
        # lr = max(0.0, cosine_lr * (base_lr - end_lr) + end_lr)
        lr = max(0.0, cosine_lr * (1 - end_lr) + end_lr)
        return lr

    return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_cosine, last_epoch)


def param_groups_lrd(model, weight_decay=0.05, no_weight_decay_list=[], layer_decay=.75, base_lr=1e-3):
    """
    Parameter groups for layer-wise lr decay
    Following BEiT: https://github.com/microsoft/unilm/blob/master/beit/optim_factory.py#L58
    """
    param_group_names = {}
    param_groups = {}

    try:
        num_layers = len(model.blocks) + 1
    except:
        num_layers = len(model.layers) + 1

    layer_scales = list(layer_decay ** (num_layers - i) for i in range(num_layers + 1))

    for n, p in model.named_parameters():
        if not p.requires_grad:
            continue

        # no decay: all 1D parameters and model specific ones
        if p.ndim == 1 or n in no_weight_decay_list:
            g_decay = "no_decay"
            this_decay = 0.
        else:
            g_decay = "decay"
            this_decay = weight_decay

        layer_id = get_layer_id_for_vit(n, num_layers)
        group_name = "layer_%d_%s" % (layer_id, g_decay)

        if group_name not in param_group_names:
            this_scale = layer_scales[layer_id]

            param_group_names[group_name] = {
                "lr_scale": this_scale,
                "weight_decay": this_decay,
                "params": [],
                "lr": base_lr,
            }
            param_groups[group_name] = {
                "lr_scale": this_scale,
                "weight_decay": this_decay,
                "params": [],
                "lr": base_lr,
            }

        param_group_names[group_name]["params"].append(n)
        param_groups[group_name]["params"].append(p)

    # print("parameter groups: \n%s" % json.dumps(param_group_names, indent=2))

    return list(param_groups.values())


def get_layer_id_for_vit(name, num_layers):
    """
    Assign a parameter with its layer id
    Following BEiT: https://github.com/microsoft/unilm/blob/master/beit/optim_factory.py#L33
    """
    if name in ['cls_token', 'pos_embed']:
        return 0
    elif name.startswith('patch_embed'):
        return 0
    # todo MAE use 'blocks' and VLP use 'layers'
    elif name.startswith('layers'):
        return int(name.split('.')[1]) + 1
    else:
        return num_layers



from hcpcvlr.utils.cfgs_loader import load_yaml
from hcpcvlr.models.mrg import Baseline, VLCI
from hcpcvlr.modules.tokenizers import MRGTokenizer
from hcpcvlr.api.pipeline import MRGPipeline
from hcpcvlr.utils.metrics import MetricCalculator
from hcpcvlr.data import MRGDataLoader
from hcpcvlr.modules.losses.nlg import compute_lm_loss
from hcpcvlr.utils.optimizer import build_lr_scheduler, build_optimizer

cfgs = load_yaml("configs/mrg/baseline.yaml")
# print(cfgs)
token = MRGTokenizer(cfgs)
model = Baseline(cfgs, token)
# -------------------
# inference
# -------------------
work = MRGPipeline(model, cfgs, metric_caculator=MetricCalculator(cfgs))
test_dataloader = MRGDataLoader(cfgs, token, split='test', shuffle=False)
work.inference(test_dataloader)
# -------------------
# training
# -------------------
cfgs = load_yaml("configs/mrg/vlci.yaml")
token = MRGTokenizer(cfgs)
model = VLCI(cfgs, token)
test_dataloader = MRGDataLoader(cfgs, token, split='test', shuffle=False)
train_dataloader = MRGDataLoader(cfgs, token, split='train', shuffle=True)
val_dataloader = MRGDataLoader(cfgs, token, split='val', shuffle=False)

optimizer = build_optimizer(cfgs, model)
lr_scheduler = build_lr_scheduler(cfgs, optimizer, len(train_dataloader))

work = MRGPipeline(model, cfgs, 
                criterion=compute_lm_loss, 
                metric_caculator=MetricCalculator(cfgs),
                optimizer=optimizer,
                lr_scheduler=lr_scheduler,
                train_dataloader=train_dataloader,
                val_dataloader=val_dataloader,
                test_dataloader=test_dataloader
                )

work.train()

from hcpcvlr.utils.cfgs_loader import load_yaml
from hcpcvlr.models.mrg import Baseline
from hcpcvlr.modules.tokenizers import MRGTokenizer
from hcpcvlr.api.pipeline import MRGPipeline
from hcpcvlr.utils.metrics import MetricCalculator
from hcpcvlr.data import MRGDataLoader


cfgs = load_yaml("configs/mrg/baseline.yaml")
token = MRGTokenizer(cfgs)
model = Baseline(cfgs, token)

# -------------------
# inference
# -------------------
work = MRGPipeline(model, cfgs, metric_caculator=MetricCalculator(cfgs))
test_dataloader = MRGDataLoader(cfgs, token, split='test', shuffle=False)
work.inference(test_dataloader)



import torch
import torch.nn as nn
import torch.nn.functional as F
from modules.backbones.transformer import EncoderLayer, SublayerConnection, MultiHeadedAttention, LayerNorm, PositionwiseFeedForward, clones
import math
from einops.layers.torch import Rearrange
from einops import rearrange


class AF(nn.Module):
    """
    Attention Fuse Module
    """
    def __init__(self, embed_dim):
        super(AF, self).__init__()
        self.embed_dim = embed_dim
        self.q = nn.Linear(embed_dim, embed_dim)
        self.out = nn.Linear(embed_dim, embed_dim)

    def forward(self, q, k, v, proj=False):
        if proj:
            q = self.q(q)
            qk = torch.matmul(q, k.transpose(-1, -2)) / math.sqrt(self.embed_dim)
            score = qk.softmax(dim=-1)
            out = score.matmul(v)
            out = self.out(out)
        else:
            qk = torch.matmul(q, k.transpose(-1, -2))
            score = qk.softmax(dim=-1)
            out = score.matmul(v)
        return out


class FDIntervention(nn.Module):
    """
    Front-Door Intervention Module
    """
    def __init__(self, embed_dim):
        super(FDIntervention, self).__init__()
        self.embed_dim = embed_dim
        self.af_1 = AF(embed_dim)
        self.af_2 = AF(embed_dim)

    def forward(self, feature, mediator, proj=False):
        v = self.af_1(mediator, feature, feature, proj)
        out = self.af_2(feature, mediator, v, proj)
        return out


class LGFM(nn.Module):
    """
    Local-Global Fuse Module
    """
    def __init__(self, embed_dim):
        super(LGFM, self).__init__()
        self.embed_dim = embed_dim
        self.norm_l = LayerNorm(embed_dim)
        self.norm_g = LayerNorm(embed_dim)
        self.gelu = nn.GELU()
        self.llf = MultiHeadedAttention(8, embed_dim)
        self.lgf = MultiHeadedAttention(8, embed_dim)
        self.proj = nn.Linear(embed_dim * 2, embed_dim)
        self.norm = LayerNorm(embed_dim)
        self.fc = nn.Sequential(
            nn.Linear(embed_dim, embed_dim * 4),
            nn.Linear(embed_dim * 4, embed_dim),
        )

    def forward(self, fl, fg):
        fl = self.norm_l(fl)
        fg = self.norm_g(fg)
        fll = fl + self.llf(fl, fl, fl)
        flg = fl + self.lgf(fl, fg, fg)
        out = self.proj(torch.cat([fll, flg], dim=-1))
        out = self.norm(out)
        out = out + self.fc(out)
        return out


class CrossLayer(EncoderLayer):
    """
    composed by a cross-attention layer and feed-forward network
    """
    def __init__(self, embed_dim, num_heads, ff_dim, dropout):
        super(CrossLayer, self).__init__(embed_dim, num_heads, ff_dim, dropout)

    def forward(self, x, y):
        x = self.sublayer_attn(x, lambda x: self.attn(x, y, y, None))
        x = self.sublayer_ff(x, self.feed_forward)
        return x


class PartAttention(nn.Module):
    """
    @article{he2021transfg,
    title={TransFG: A Transformer Architecture for Fine-grained Recognition},
    author={He, Ju and Chen, Jie-Neng and Liu, Shuai and Kortylewski, Adam and Yang, Cheng and Bai, Yutong and Wang, Changhu and Yuille, Alan},
    journal={arXiv preprint arXiv:2103.07976},
    year={2021}
    }
    """
    def __init__(self):
        super(PartAttention, self).__init__()

    def forward(self, x, k=6):
        """
        x -> list: the attention list from the encoder
        k: select the top k attention score and their index from each head
        """
        length = len(x)
        last_map = x[0]
        for i in range(1, length):
            last_map = torch.matmul(x[i], last_map)
        last_map = last_map[:, :, 0, 1:]
        max_attn, max_inx = last_map.sort(2, descending=True)
        max_inx = max_inx[:, :, :k].reshape([last_map.size(0), -1])
        max_attn = max_attn[:, :, :k].reshape([last_map.size(0), -1])
        return max_attn, max_inx


class RelativeAttention2D(nn.Module):
    def __init__(self, inp, oup, image_size, heads=8, dim_head=32, dropout=0.):
        super().__init__()
        inner_dim = dim_head * heads
        project_out = not (heads == 1 and dim_head == inp)

        self.ih, self.iw = image_size

        self.heads = heads
        self.scale = dim_head ** -0.5

        # parameter table of relative position bias
        self.relative_bias_table = nn.Parameter(
            torch.zeros((2 * self.ih - 1) * (2 * self.iw - 1), heads))

        coords = torch.meshgrid((torch.arange(self.ih), torch.arange(self.iw)))
        coords = torch.flatten(torch.stack(coords), 1)
        relative_coords = coords[:, :, None] - coords[:, None, :]

        relative_coords[0] += self.ih - 1
        relative_coords[1] += self.iw - 1
        relative_coords[0] *= 2 * self.iw - 1
        relative_coords = rearrange(relative_coords, 'c h w -> h w c')
        relative_index = relative_coords.sum(-1).flatten().unsqueeze(1)
        self.register_buffer("relative_index", relative_index)

        self.attend = nn.Softmax(dim=-1)
        self.to_qkv = nn.Linear(inp, inner_dim * 3, bias=False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, oup),
            nn.Dropout(dropout)
        ) if project_out else nn.Identity()

    def forward(self, x):
        qkv = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(
            t, 'b n (h d) -> b h n d', h=self.heads), qkv)

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        # Use "gather" for more efficiency on GPUs
        relative_bias = self.relative_bias_table.gather(
            0, self.relative_index.repeat(1, self.heads))
        relative_bias = rearrange(
            relative_bias, '(h w) c -> 1 c h w', h=self.ih * self.iw, w=self.ih * self.iw)
        dots = dots + relative_bias

        attn = self.attend(dots)
        out = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        out = self.to_out(out)
        return out


class LocalSample(nn.Module):
    def __init__(self, embed_dim, num_heads, ff_dim, dropout=0.):
        super(LocalSample, self).__init__()
        self.part_select = PartAttention()
        self.causal_layer = CaaM(embed_dim, num_heads, ff_dim, dropout)

    def forward(self, x, attn, k):
        """
        x: all visual tokens
        attn: attention from encoder
        k: select k local feature

        fl: k local feature without [CLS] token
        """
        part_attn, part_inx = self.part_select(attn, k)
        part_inx = part_inx + 1
        parts = []
        B, num = part_inx.shape
        for i in range(B):
            parts.append(x[i, part_inx[i, :]])
        fl = torch.stack(parts).squeeze(1)
        fl = torch.cat([x[:, :1], fl], dim=1)
        fl = self.causal_layer(fl)[:, 1:]
        return fl


class PreNorm(nn.Module):
    def __init__(self, dim, fn, norm):
        super().__init__()
        self.norm = norm(dim)
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)


class DownSamplingTrans(nn.Module):
    """
    @article{dai2021coatnet,
    title={CoAtNet: Marrying Convolution and Attention for All Data Sizes},
    author={Dai, Zihang and Liu, Hanxiao and Le, Quoc V and Tan, Mingxing},
    journal={arXiv preprint arXiv:2106.04803},
    year={2021}
    }
    """
    def __init__(self, inp, oup, image_size, heads=8, dim_head=32, downsample=False, dropout=0.):
        super().__init__()
        hidden_dim = int(inp * 4)

        self.ih, self.iw = image_size
        self.downsample = downsample

        if self.downsample:
            self.pool1 = nn.MaxPool2d(3, 2, 1)
            self.pool2 = nn.MaxPool2d(3, 2, 1)
            self.proj = nn.Conv2d(inp, oup, 1, 1, 0, bias=False)

        self.attn = RelativeAttention2D(inp, oup, image_size, heads, dim_head, dropout)
        self.ff = PositionwiseFeedForward(oup, hidden_dim, dropout)

        self.attn = nn.Sequential(
            Rearrange('b c ih iw -> b (ih iw) c'),
            PreNorm(inp, self.attn, nn.LayerNorm),
            Rearrange('b (ih iw) c -> b c ih iw', ih=self.ih, iw=self.iw)
        )

        self.ff = nn.Sequential(
            Rearrange('b c ih iw -> b (ih iw) c'),
            PreNorm(oup, self.ff, nn.LayerNorm),
            Rearrange('b (ih iw) c -> b c ih iw', ih=self.ih, iw=self.iw)
        )

    def forward(self, x):
        if self.downsample:
            x = self.proj(self.pool1(x)) + self.attn(self.pool2(x))
        else:
            x = x + self.attn(x)
        x = x + self.ff(x)
        return x


class GlobalSample(nn.Module):
    def __init__(self, embed_dim, num_heads, ff_dim, dropout):
        super(GlobalSample, self).__init__()
        self.global_sample = DownSamplingTrans(embed_dim, embed_dim, (7, 7), downsample=True)

    def forward(self, x):
        img = x[:, 1:, :]
        B, L, N = img.size()
        img = img.reshape([-1, 14, 14, N])

        img = img.permute(0, 3, 1, 2)
        img = self.global_sample(img)
        img = img.permute(0, 2, 3, 1)

        fg = img.reshape([B, -1, N])
        return fg


class CausalAttention(nn.Module):
    def __init__(self, h, d_model, dropout=0.1):
        super(CausalAttention, self).__init__()
        assert d_model % h == 0
        self.d_k = d_model // h
        self.h = h
        self.linears = clones(nn.Linear(d_model, d_model), 4)
        self.attn = None
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x):
        q = x
        k = x
        v = x

        nbatches = q.size(0)
        q, k, v = \
            [l(x).view(nbatches, -1, self.h, self.d_k).transpose(1, 2)
             for l, x in zip(self.linears, (q, k, v))]

        d_k = q.size(-1)

        scores = torch.matmul(q, k.transpose(-2, -1)) / math.sqrt(d_k)
        p_attn = F.softmax(-scores, dim=-1)
        p_attn_comp = F.softmax(scores, dim=-1)

        self.attn = p_attn
        p_attn = self.dropout(p_attn)
        p_attn_comp = self.dropout(p_attn_comp)
        out = torch.matmul(p_attn, v)
        out = out.transpose(1, 2).contiguous().view(nbatches, -1, self.h * self.d_k)
        out = self.linears[-1](out)
        out_comp = torch.matmul(p_attn_comp, v)
        out_comp = out_comp.transpose(1, 2).contiguous().view(nbatches, -1, self.h * self.d_k)
        out_comp = self.linears[-1](out_comp)
        return out, out_comp


class CaaMSublayerConnection(nn.Module):
    def __init__(self, embed_dim, dropout):
        super(CaaMSublayerConnection, self).__init__()
        self.norm = LayerNorm(embed_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, sublayer):
        attn, attn_comp = sublayer(self.norm(x))
        attn = self.dropout(attn)
        attn_comp = self.dropout(attn_comp)
        return x + attn, attn_comp


class CaaM(nn.Module):
    """
    @inproceedings{wang2021causal,
    title={Causal Attention for Unbiased Visual Recognition},
    author={Wang, Tan and Zhou, Chang and Sun, Qianru and Zhang, Hanwang},
    booktitle={Proceedings of the IEEE/CVF International Conference on Computer Vision (ICCV)},
    year={2021}
    }
    """
    def __init__(self, embed_dim, num_heads, ff_dim, dropout):
        super(CaaM, self).__init__()
        self.attn = CausalAttention(num_heads, embed_dim, dropout)
        self.feed_forward = PositionwiseFeedForward(embed_dim, ff_dim, dropout)
        self.sublayer_attn = CaaMSublayerConnection(embed_dim, dropout)
        self.sublayer_ff = SublayerConnection(embed_dim, dropout)

    def forward(self, x, mask=None):
        x, x_comp = self.sublayer_attn(x, lambda x: self.attn(x))
        x = self.sublayer_ff(x, self.feed_forward)
        x = x + self.sublayer_ff(x_comp, self.feed_forward)
        return x


class VDM(nn.Module):
    def __init__(self, embed_dim, num_heads, ff_dim, dropout=.1):
        super(VDM, self).__init__()
        self.fuse = LGFM(embed_dim)
        self.intervene = FDIntervention(embed_dim)

    def forward(self, x, fl=None, fg=None, mode=True, proj=False):
        """
        fl: local feature
        fg: global feature
        """
        if mode:
            mt = self.fuse(fl, fg)
            out = self.intervene(x, mt, proj)
        else:
            out = 0
        return out + x


class LDM(nn.Module):
    def __init__(self, embed_dim):
        super(LDM, self).__init__()
        self.embed_dim = embed_dim
        self.fuse_v = CrossLayer(embed_dim, 8, 4*embed_dim, .1)
        self.fuse_t = CrossLayer(embed_dim, 8, 4*embed_dim, .1)
        self.norm = LayerNorm(embed_dim)
        self.intervene = FDIntervention(embed_dim)

    def forward(self, x, text, vis, mode=True, proj=False):
        if mode:
            mediator = self.fuse_t(vis, text)
            mediator = self.fuse_v(mediator, vis) + mediator
            mediator = self.norm(mediator)
            out = self.intervene(x, mediator, proj)
        else:
            out = 0
        return out + x



from .nlg.bleu.bleu import Bleu
from .nlg.meteor import Meteor
from .nlg.rouge import Rouge
from .nlg.cider.cider import Cider


METRIC_LIST = {
    # NLG
    "Meteor": Meteor,
    "Rouge": Rouge,
    "Cider": Cider,
    "Bleu": Bleu
}


class MetricCalculator():
    def __init__(self, cfgs):
        # TODO
        self.metric_name = cfgs["metric"]
    
    def get_scorer(self):
        scorers = []
        for key in self.metric_name:
            if key == "Bleu":
                scorers.append((Bleu(4), ["BLEU_1", "BLEU_2", "BLEU_3", "BLEU_4"]))    
            else:
                scorers.append((METRIC_LIST[key](), key))
        return scorers
    
    def compute_scores(self, gts, res):
        """
        Performs the MS COCO evaluation using the Python 3 implementation (https://github.com/salaniz/pycocoevalcap)

        :param gts: Dictionary with the image ids and their gold captions,
        :param res: Dictionary with the image ids ant their generated captions
        :print: Evaluation score (the mean of the scores of all the instances) for each measure
        """
        # Set up scorers
        scorers = self.get_scorer()
        eval_res = {}
        # Compute score for each metric
        for scorer, method in scorers:
            try:
                score, scores = scorer.compute_score(gts, res, verbose=0)
            except TypeError:
                score, scores = scorer.compute_score(gts, res)
            if type(method) == list:
                for sc, m in zip(score, method):
                    eval_res[m] = sc
            else:
                eval_res[method] = score
        return eval_res



import torch
from torch import nn
from torch.autograd import forward_ad
from modules.backbones.transformer import MultiHeadedAttention, PositionwiseFeedForward, LayerNorm
from torchvision.models import resnet101
from ..utils.pos_embed import get_2d_sincos_pos_embed


class PatchEmbed(nn.Module):
    """
    TODO add more selections for patch embedding
    resnet 1-3 block stem
    """

    def __init__(self, img_size=224, patch_size=16):
        super(PatchEmbed, self).__init__()
        img_size = (img_size, img_size)
        patch_size = (patch_size, patch_size)
        num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0])
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        model = resnet101(True)
        modules = list(model.children())[:-3]
        self.embed = nn.Sequential(*modules)

    def forward(self, x):
        B, C, H, W = x.shape
        # FIXME look at relaxing size constraints
        assert H == self.img_size[0] and W == self.img_size[1], \
            f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]})."
        x = self.embed(x).flatten(2).transpose(1, 2)
        return x


class VisEmbed(nn.Module):
    """
    image embedding with 2d sin-cos position embedding
    """
    def __init__(self, img_size=224, patch_size=16, embed_dim=512, dropout=0., cls_token=True):
        super(VisEmbed, self).__init__()

        # --------------------------------------------------------------------------
        # Patchfy and Embedding
        self.patch_embed = PatchEmbed(img_size, patch_size)
        self.dropout = nn.Dropout(p=dropout)
        self.proj = nn.Linear(1024, embed_dim)
        self.norm = LayerNorm(embed_dim)
        num_patches = self.patch_embed.num_patches
        # use 2d pos embed
        self.cls_token = cls_token
        if cls_token:
            self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim),
                                      requires_grad=False)  # fixed sin-cos embedding
        else:
            self.pos_embed = nn.Parameter(torch.zeros(1, num_patches, embed_dim),
                                      requires_grad=False)  # fixed sin-cos embedding

        # self.norm = norm_layer(embed_dim)
        self.initialize_weights()

        # initialize nn.Linear and nn.LayerNorm
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def initialize_weights(self):
        # initialization
        pos_embed = get_2d_sincos_pos_embed(self.pos_embed.shape[-1], int(self.patch_embed.num_patches ** .5),
                                            cls_token=self.cls_token)
        self.pos_embed.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))

    def forward(self, x):
        x = self.patch_embed(x)
        if self.cls_token:
            x = self.proj(x) + self.pos_embed[:, 1:, :]
        else:
            x = self.proj(x) + self.pos_embed
        x = self.dropout(self.norm(x))
        return x


class VisionTransformer(nn.Module):
    def __init__(self):
        super().__init__()
    
    def _init_weight(self):
        pass

    def forward(self):
        pass



import yaml


def load_yaml(yaml_path, cfgs=None):
    """
    loading config file .yaml
    and extend the dict.
    """
    if cfgs is None:
        cfgs = {}
    with open(yaml_path, 'r', encoding='utf-8') as f:
        _cfgs = yaml.load(f.read(), Loader=yaml.FullLoader)
        for key in _cfgs.keys():
            if key == '__base__':
                cfgs = load_yaml(_cfgs[key], cfgs)
            else:
                cfgs.update(_cfgs[key])
    return cfgs

def load_json(json_path):
    pass



def get_question_text(problem):
    question = problem['question']
    return question


def get_context_text(problem, use_caption):
    txt_context = problem['hint']
    img_context = problem['caption'] if use_caption else ""
    context = " ".join([txt_context, img_context]).strip()
    if context == "":
        context = "N/A"
    return context


def get_choice_text(probelm, options):
    choices = probelm['choices']
    choice_list = []
    for i, c in enumerate(choices):
        choice_list.append("({}) {}".format(options[i], c))
    choice_txt = " ".join(choice_list)
    #print(choice_txt)
    return choice_txt


def get_answer(problem, options):
    return options[problem['answer']]


def get_lecture_text(problem):
    # \\n: GPT-3 can generate the lecture with more tokens.
    lecture = problem['lecture'].replace("\n", "\\n")
    return lecture


def get_solution_text(problem):
    # \\n: GPT-3 can generate the solution with more tokens
    solution = problem['solution'].replace("\n", "\\n")
    return solution


def create_one_example(format, question, context, choice, answer, lecture, solution, test_example=True):

    input_format, output_format = format.split("-")

    ## Inputs
    if input_format == "CQM":
        input = f"Context: {context}\nQuestion: {question}\nOptions: {choice}\n"
    elif input_format == "QCM":
        input = f"Question: {question}\nContext: {context}\nOptions: {choice}\n"
    # upper bound experiment
    elif input_format == "QCML":
        input = f"Question: {question}\nContext: {context}\nOptions: {choice}\nBECAUSE: {lecture}\n"
    elif input_format == "QCME":
        input = f"Question: {question}\nContext: {context}\nOptions: {choice}\nBECAUSE: {solution}\n"
    elif input_format == "QCMLE":
        input = f"Question: {question}\nContext: {context}\nOptions: {choice}\nBECAUSE: {lecture} {solution}\n"

    elif input_format == "QCLM":
        input = f"Question: {question}\n\nContext: \"{context}\"\n\nLecture: \"{lecture}\"\n\nOptions: {choice}\n"
    elif input_format == "QCEM":
        input = f"Question: {question}\nContext: {context}\nBECAUSE: {solution}\nOptions: {choice}\n"
    elif input_format == "QCLEM":
        input = f"Question: {question}\nContext: {context}\nBECAUSE: {lecture} {solution}\nOptions: {choice}\n"

    # Outputs
    if test_example:
        output = "" # Answer: (Provide your answer with tags, e.g., <Answer>A/B/C</Answer>.)
    elif output_format == 'A':
        output = f"Answer: The answer is <Answer>{answer}</Answer>."

    elif output_format == 'AL':
        output = f"Answer: The answer is <Answer>{answer}</Answer>. BECAUSE: {solution}"
    elif output_format == 'AE':
        output = f"Answer: The answer is <Answer>{answer}</Answer>. BECAUSE: {lecture}"
    elif output_format == 'ALE':
        output = f"Answer: The answer is <Answer>{answer}</Answer>. BECAUSE: {lecture} {solution}"
    elif output_format == 'AEL':
        output = f"Answer: The answer is <Answer>{answer}</Answer>. BECAUSE: {solution} {lecture}"

    elif output_format == 'LA':
        output = f"Answer: {lecture} The answer is <Answer>{answer}</Answer>."
    elif output_format == 'EA':
        output = f"Answer: {solution} Therefore, the answer is <Answer>{answer}</Answer>."
    elif output_format == 'LEA':
        output = f"Answer: {lecture} {solution} The answer is <Answer>{answer}</Answer>."
    elif output_format == 'ELA':
        output = f"Answer: {solution} {lecture} The answer is <Answer>{answer}</Answer>."
    elif output_format == 'CoTEA':
        output = f"Answer: Let's think step by step. {solution} The answer is <Answer>{answer}</Answer>."

    text = input + output
    text = text.replace("  ", " ").strip()
    if text.endswith("BECAUSE:"):
        text = text.replace("BECAUSE:", "").strip()
    return text


def build_prompt(problems, shot_qids, test_qid, args):

    examples = []
    

    # n-shot training examples
    for qid in shot_qids:
        # print(problems[qid]['caption'])
        # ['question', 'choices', 'answer', 'hint', 'image', 'task', 'grade', 'subject', 'topic', 'category', 'skill', 'lecture', 'solution', 'split', 'caption']
        question = get_question_text(problems[qid])
        context = get_context_text(problems[qid], args.use_caption)
        choice = get_choice_text(problems[qid], args.options)
        answer = get_answer(problems[qid], args.options)
        lecture = get_lecture_text(problems[qid])
        solution = get_solution_text(problems[qid])

        train_example = create_one_example(args.prompt_format,
                                           question,
                                           context,
                                           choice,
                                           answer,
                                           lecture,
                                           solution,
                                           test_example=False)
        examples.append(train_example)

    # test example
    question = get_question_text(problems[test_qid])
    context = get_context_text(problems[test_qid], args.use_caption)
    choice = get_choice_text(problems[test_qid], args.options)
    answer = get_answer(problems[test_qid], args.options)
    lecture = get_lecture_text(problems[test_qid])
    solution = get_solution_text(problems[test_qid])

    test_example = create_one_example(args.prompt_format,
                                      question,
                                      context,
                                      choice,
                                      answer,
                                      lecture,
                                      solution,
                                      test_example=True)
    examples.append(test_example)

    # create the prompt input
    prompt_input = '\n\n'.join(examples)

    return prompt_input


import torch
from torch import nn
from modules.prompt.caco_cot import *
import requests
from .cot import CoT
from time import sleep
import numpy as np


class CaCoCoT(CoT):
    def __init__(self, cfgs):
        super().__init__(cfgs)
    

    # @retry(delay=3, tries=10, backoff=2, max_delay=120)
    def answer_review(self, question=None, choice=None, depth=1, history=''):
        get_single_run = self.get_single_run_gpt

        output_1 = get_single_run(None, our_reasoner1_prompt + question)
        answer_1 = self.filter_output(output_1, choice=choice)

        sleep(0.1)
        output_2 = get_single_run(None, our_reasoner2_prompt + question)
        answer_2 = self.filter_output(output_2, choice=choice)

        history += f" ########## Round {depth} ########## \n" + str(output_1 + '\n\n' + '-'*30 + '\n\n' + output_2 + '\n\n')

        answers = []

        if answer_1 == answer_2: 

            lst = ['A', 'B', 'C', 'D', 'E']
            lst = lst[:len(choice)]
            lst.pop(lst.index(answer_1))

            answer_cf = np.random.choice(lst)
            reviewer_prompt1 = our_reviewer_prompt + f"Now, For the question: \n\"\"\"\n{question}\n\"\"\" \n\n\nA possible solution is: \n\"\"\"\n{np.random.choice([output_1, output_2])}\n\"\"\"\n\nPlease provide your evaluation, while looking out for option ({answer_cf}). Remember, if the answer cannot be determined, make an educated guess at the end. "

            reviewer_output_3 = get_single_run(None, reviewer_prompt1)
            answer_3 = self.filter_output(reviewer_output_3, choice)

            history += f" ######## Round {depth} Review ######## \n\n" + str(reviewer_output_3 + '\n\n' + '='*30 + '\n\n')

            answers = [answer_1, answer_2, answer_3]
            max_answer = max(set(answers), key=answers.count)

            if answer_3 == answer_1: 
                history += f'\n\n{answer_1} {answer_2} {answer_3} CF: {answer_cf}  >>> all consensus \n\n'
                return answer_3, history
            elif depth >= 4: 
                history += f'\n\n{answer_1} {answer_2} {answer_3} CF: {answer_cf} >>> reach max depth\n\n'
                return np.random.choice([answer_1, answer_2,]), history
            else: 
                history += f'\n\n{answer_1} {answer_2} {answer_3} CF: {answer_cf} >>> go deeper\n\n'
                return self.answer_review(question=question, choice=choice, depth=depth+1, history=history)

        elif answer_1 != answer_2: 

            reviewer_prompt1 = our_reviewer_prompt + f"Now, For the question: \n\"\"\"\n{question}\n\"\"\" \n\n\nA possible solution is: \n\"\"\"\n{output_1}\n\"\"\"\n\nPlease provide your evaluation, while looking out for option ({answer_2}). Remember, if the answer cannot be determined, make an educated guess at the end. "
            reviewer_prompt2 = our_reviewer_prompt + f"Now, For the question: \n\"\"\"\n{question}\n\"\"\" \n\n\nA possible solution is: \n\"\"\"\n{output_2}\n\"\"\"\n\nPlease provide your evaluation, while looking out for option ({answer_1}). Remember, if the answer cannot be determined, make an educated guess at the end. "

            reviewer_output_3 = get_single_run(None, reviewer_prompt1)
            reviewer_output_4 = get_single_run(None, reviewer_prompt2)
            
            answer_3 = self.filter_output(reviewer_output_3, choice)
            answer_4 = self.filter_output(reviewer_output_4, choice)

            history += f" ######## Round {depth} Review ######## \n\n" + str(reviewer_output_3 + '\n\n' + '-'*30 + '\n\n' + reviewer_output_4 + '\n\n' + '='*30 + '\n\n')

            answers = [answer_1, answer_2, answer_3, answer_4]
            max_answer = max(set(answers), key=answers.count)

            if answer_3 == answer_4: 
                history += f'\n\n{answer_1} {answer_2} {answer_3} {answer_4} >>> evaluators\' consensus \n\n'
                return answer_3, history
            elif depth >= 4: 
                history += f'\n\n{answer_1} {answer_2} {answer_3} {answer_4} >>> reach the max depth\n\n'
                return np.random.choice([answer_1, answer_2]), history
            else: 
                history += f'\n\n{answer_1} {answer_2} {answer_3} {answer_4} >>> go deeper\n\n'
                return self.answer_review(question=question, choice=choice, depth=depth+1, history=history)


    def __call__(self, qid, prompt, choice):
        get_single_run = self.get_single_run_gpt
        
        answer, output = self.answer_review(question=prompt, choice=choice, depth=1, history='')
            
        return qid, answer, output


__author__ = 'tylin'

from .tokenizer.ptbtokenizer import PTBTokenizer
from .bleu.bleu import Bleu
from .meteor.meteor import Meteor
from .rouge.rouge import Rouge
from .cider.cider import Cider


class COCOEvalCap:
    def __init__(self, coco, cocoRes):
        self.evalImgs = []
        self.eval = {}
        self.imgToEval = {}
        self.coco = coco
        self.cocoRes = cocoRes
        self.params = {'image_id': cocoRes.getImgIds()}

    def evaluate(self):
        imgIds = self.params['image_id']
        # imgIds = self.coco.getImgIds()
        gts = {}
        res = {}
        for imgId in imgIds:
            gts[imgId] = self.coco.imgToAnns[imgId]
            res[imgId] = self.cocoRes.imgToAnns[imgId]

        # =================================================
        # Set up scorers
        # =================================================
        print('tokenization...')
        tokenizer = PTBTokenizer()
        gts = tokenizer.tokenize(gts)
        res = tokenizer.tokenize(res)

        # =================================================
        # Set up scorers
        # =================================================
        print('setting up scorers...')
        scorers = [
            (Bleu(4), ["Bleu_1", "Bleu_2", "Bleu_3", "Bleu_4"]),
            (Meteor(), "METEOR"),
            (Rouge(), "ROUGE_L"),
            (Cider(), "CIDEr")
        ]

        # =================================================
        # Compute scores
        # =================================================
        eval = {}
        for scorer, method in scorers:
            print('computing %s score...' % (scorer.method()))
            score, scores = scorer.compute_score(gts, res)
            if type(method) == list:
                for sc, scs, m in zip(score, scores, method):
                    self.setEval(sc, m)
                    self.setImgToEvalImgs(scs, imgIds, m)
                    print("%s: %0.3f" % (m, sc))
            else:
                self.setEval(score, method)
                self.setImgToEvalImgs(scores, imgIds, method)
                print("%s: %0.3f" % (method, score))
        self.setEvalImgs()

    def setEval(self, score, method):
        self.eval[method] = score

    def setImgToEvalImgs(self, scores, imgIds, method):
        for imgId, score in zip(imgIds, scores):
            if not imgId in self.imgToEval:
                self.imgToEval[imgId] = {}
                self.imgToEval[imgId]["image_id"] = imgId
            self.imgToEval[imgId][method] = score

    def setEvalImgs(self):
        self.evalImgs = [eval for imgId, eval in self.imgToEval.items()]



import torch
import numpy as np


def penalty_builder(penalty_config):
    if penalty_config == '':
        return lambda x, y: y
    pen_type, alpha = penalty_config.split('_')
    alpha = float(alpha)
    if pen_type == 'wu':
        return lambda x, y: length_wu(x, y, alpha)
    if pen_type == 'avg':
        return lambda x, y: length_average(x, y, alpha)


def length_wu(length, logprobs, alpha=0.):
    """
    NMT length re-ranking score from
    "Google's Neural Machine Translation System" :cite:`wu2016google`.
    """

    modifier = (((5 + length) ** alpha) /
                ((5 + 1) ** alpha))
    return logprobs / modifier


def length_average(length, logprobs, alpha=0.):
    """
    Returns the average probability of tokens in a sequence.
    """
    return logprobs / length


def split_tensors(n, x):
    if torch.is_tensor(x):
        assert x.shape[0] % n == 0
        x = x.reshape(x.shape[0] // n, n, *x.shape[1:]).unbind(1)
    elif type(x) is list or type(x) is tuple:
        x = [split_tensors(n, _) for _ in x]
    elif x is None:
        x = [None] * n
    return x


def repeat_tensors(n, x):
    """
    For a tensor of size Bx..., we repeat it n times, and make it Bnx...
    For collections, do nested repeat
    """
    if torch.is_tensor(x):
        x = x.unsqueeze(1)  # Bx1x...
        x = x.expand(-1, n, *([-1] * len(x.shape[2:])))  # Bxnx...
        x = x.reshape(x.shape[0] * n, *x.shape[2:])  # Bnx...
    elif type(x) is list or type(x) is tuple:
        x = [repeat_tensors(n, _) for _ in x]
    return x


def subsequent_mask(size):
    attn_shape = (1, size, size)
    subsequent_mask = np.triu(np.ones(attn_shape), k=1).astype('uint8')
    return torch.from_numpy(subsequent_mask) == 0



"""
Tokenizer for IU-Xray, MIMIC-CXR and FFA-IR dataset
"""
import json
import re
from collections import Counter


class Tokenizer(object):
    def __init__(self, cfgs):
        self.ann_path = cfgs["ann_path"]
        self.threshold = cfgs["threshold"]
        self.dataset_name = cfgs["dataset_name"]
        if self.dataset_name == 'iu_xray':
            self.clean_report = self.clean_report_iu_xray
        elif self.dataset_name == 'mimic_cxr':
            self.clean_report = self.clean_report_mimic_cxr
        elif self.dataset_name == 'ffa_ir':
            self.clean_report = self.clean_report_ffa_ir
        self.ann = json.loads(open(self.ann_path, 'r').read())
        self.token2idx, self.idx2token = self.create_vocabulary()

    def create_vocabulary(self):
        total_tokens = []
        if self.dataset_name == 'ffa_ir':
            _report_key = 'En_Report'
            for example in self.ann['train']:
                tokens = self.clean_report(self.ann['train'][example][_report_key]).split()
                for token in tokens:
                    total_tokens.append(token)
        else:
            _report_key = 'report'
            for example in self.ann['train']:
                tokens = self.clean_report(example[_report_key]).split()
                for token in tokens:
                    total_tokens.append(token)

        counter = Counter(total_tokens)
        vocab = [k for k, v in counter.items() if v >= self.threshold] + ['<unk>']
        vocab.sort()
        token2idx, idx2token = {}, {}
        for idx, token in enumerate(vocab):
            token2idx[token] = idx + 1
            idx2token[idx + 1] = token
        return token2idx, idx2token

    def clean_report_iu_xray(self, report):
        report_cleaner = lambda t: t.replace('..', '.').replace('..', '.').replace('..', '.').replace('1. ', '') \
            .replace('. 2. ', '. ').replace('. 3. ', '. ').replace('. 4. ', '. ').replace('. 5. ', '. ') \
            .replace(' 2. ', '. ').replace(' 3. ', '. ').replace(' 4. ', '. ').replace(' 5. ', '. ') \
            .strip().lower().split('. ')
        sent_cleaner = lambda t: re.sub('[.,?;*!%^&_+():-\[\]{}]', '', t.replace('"', '').replace('/', '').
                                        replace('\\', '').replace("'", '').strip().lower())
        tokens = [sent_cleaner(sent) for sent in report_cleaner(report) if sent_cleaner(sent) != []]
        report = ' . '.join(tokens) + ' .'
        return report

    def clean_report_mimic_cxr(self, report):
        report_cleaner = lambda t: t.replace('\n', ' ').replace('__', '_').replace('__', '_').replace('__', '_') \
            .replace('__', '_').replace('__', '_').replace('__', '_').replace('__', '_').replace('  ', ' ') \
            .replace('  ', ' ').replace('  ', ' ').replace('  ', ' ').replace('  ', ' ').replace('  ', ' ') \
            .replace('..', '.').replace('..', '.').replace('..', '.').replace('..', '.').replace('..', '.') \
            .replace('..', '.').replace('..', '.').replace('..', '.').replace('1. ', '').replace('. 2. ', '. ') \
            .replace('. 3. ', '. ').replace('. 4. ', '. ').replace('. 5. ', '. ').replace(' 2. ', '. ') \
            .replace(' 3. ', '. ').replace(' 4. ', '. ').replace(' 5. ', '. ') \
            .strip().lower().split('. ')
        sent_cleaner = lambda t: re.sub('[.,?;*!%^&_+():-\[\]{}]', '', t.replace('"', '').replace('/', '')
                                        .replace('\\', '').replace("'", '').strip().lower())
        tokens = [sent_cleaner(sent) for sent in report_cleaner(report) if sent_cleaner(sent) != []]
        report = ' . '.join(tokens) + ' .'
        return report

    def clean_report_ffa_ir(self, report):
        report_cleaner = lambda t: t.replace('\n', ' ').replace('__', '_').replace('__', '_').replace('__', '_') \
            .replace('__', '_').replace('__', '_').replace('__', '_').replace('__', '_').replace('  ', ' ') \
            .replace('  ', ' ').replace('  ', ' ').replace('  ', ' ').replace('  ', ' ').replace('  ', ' ') \
            .replace('..', '.').replace('..', '.').replace('..', '.').replace('..', '.').replace('..', '.') \
            .replace('..', '.').replace('..', '.').replace('..', '.').replace('1. ', '').replace('. 2. ', '. ') \
            .replace('. 3. ', '. ').replace('. 4. ', '. ').replace('. 5. ', '. ').replace(' 2. ', '. ') \
            .replace(' 3. ', '. ').replace(' 4. ', '. ').replace(' 5. ', '. ') \
            .strip().lower().split('. ')
        sent_cleaner = lambda t: re.sub('[.,?;*!%^&_+():-\[\]{}]', '', t.replace('"', '').replace('/', '')
                                        .replace('\\', '').replace("'", '').strip().lower())
        # jieba cut the tokens
        tokens = [sent_cleaner(sent) for sent in report_cleaner(report) if sent_cleaner(sent) != []]
        report = ' . '.join(tokens) + ' .'
        return report

    def get_token_by_id(self, id):
        return self.idx2token[id]

    def get_id_by_token(self, token):
        if token not in self.token2idx:
            return self.token2idx['<unk>']
        return self.token2idx[token]

    def get_vocab_size(self):
        return len(self.token2idx)

    def __call__(self, report):
        tokens = self.clean_report(report).split()
        ids = []
        for token in tokens:
            ids.append(self.get_id_by_token(token))
        ids = [0] + ids + [0]
        return ids

    def decode(self, ids):
        txt = ''
        for i, idx in enumerate(ids):
            if idx > 0:
                if i >= 1:
                    txt += ' '
                txt += self.idx2token[idx]
            else:
                break
        return txt

    def decode_batch(self, ids_batch):
        out = []
        for ids in ids_batch:
            out.append(self.decode(ids))
        return out


class MixTokenizer(object):
    def __init__(self, cfgs):
        self.ann_path = {'iu_xray': cfgs["ann_path"]['iu_xray'],
                         'mimic_cxr': cfgs["ann_path"]['mimic_cxr']}
        self.threshold = {'iu_xray': cfgs["threshold"]['iu_xray'],
                          'mimic_cxr': cfgs["threshold"]['mimic_cxr']}
        self.dataset_name = cfgs['dataset_name']
        self.ann = {'iu_xray': json.loads(open(self.ann_path['iu_xray'], 'r').read()),
                    'mimic_cxr': json.loads(open(self.ann_path['mimic_cxr'], 'r').read())}
        self.token2idx, self.idx2token = self.create_vocabulary()

    def create_vocabulary(self):
        total_tokens_iu_xray = []
        total_tokens_mimic_cxr = []

        for example in self.ann['iu_xray']['train']:
            tokens = self.clean_report_iu_xray(example['report']).split()
            for token in tokens:
                total_tokens_iu_xray.append(token)

        for example in self.ann['mimic_cxr']['train']:
            tokens = self.clean_report_mimic_cxr(example['report']).split()
            for token in tokens:
                total_tokens_mimic_cxr.append(token)

        counter_iu_xray = Counter(total_tokens_iu_xray)
        counter_mimic_cxr = Counter(total_tokens_mimic_cxr)
        # counter vocab which more than [threshold] times appear
        vocab = [k for k, v in counter_iu_xray.items() if v >= self.threshold['iu_xray']] + ['<unk>']
        vocab += [k for k, v in counter_mimic_cxr.items() if v >= self.threshold['mimic_cxr'] and k not in vocab]
        vocab.sort()
        token2idx, idx2token = {}, {}
        for idx, token in enumerate(vocab):
            token2idx[token] = idx + 1
            idx2token[idx + 1] = token 
        return token2idx, idx2token

    def clean_report_iu_xray(self, report):
        report_cleaner = lambda t: t.replace('..', '.').replace('..', '.').replace('..', '.').replace('1. ', '') \
            .replace('. 2. ', '. ').replace('. 3. ', '. ').replace('. 4. ', '. ').replace('. 5. ', '. ') \
            .replace(' 2. ', '. ').replace(' 3. ', '. ').replace(' 4. ', '. ').replace(' 5. ', '. ') \
            .strip().lower().split('. ')
        sent_cleaner = lambda t: re.sub('[.,?;*!%^&_+():-\[\]{}]', '', t.replace('"', '').replace('/', '').
                                        replace('\\', '').replace("'", '').strip().lower())
        tokens = [sent_cleaner(sent) for sent in report_cleaner(report) if sent_cleaner(sent) != []]
        report = ' . '.join(tokens) + ' .'
        return report

    def clean_report_mimic_cxr(self, report):
        report_cleaner = lambda t: t.replace('\n', ' ').replace('__', '_').replace('__', '_').replace('__', '_') \
            .replace('__', '_').replace('__', '_').replace('__', '_').replace('__', '_').replace('  ', ' ') \
            .replace('  ', ' ').replace('  ', ' ').replace('  ', ' ').replace('  ', ' ').replace('  ', ' ') \
            .replace('..', '.').replace('..', '.').replace('..', '.').replace('..', '.').replace('..', '.') \
            .replace('..', '.').replace('..', '.').replace('..', '.').replace('1. ', '').replace('. 2. ', '. ') \
            .replace('. 3. ', '. ').replace('. 4. ', '. ').replace('. 5. ', '. ').replace(' 2. ', '. ') \
            .replace(' 3. ', '. ').replace(' 4. ', '. ').replace(' 5. ', '. ') \
            .strip().lower().split('. ')
        sent_cleaner = lambda t: re.sub('[.,?;*!%^&_+():-\[\]{}]', '', t.replace('"', '').replace('/', '')
                                        .replace('\\', '').replace("'", '').strip().lower())
        tokens = [sent_cleaner(sent) for sent in report_cleaner(report) if sent_cleaner(sent) != []]
        report = ' . '.join(tokens) + ' .'
        return report

    def get_token_by_id(self, id):
        return self.idx2token[id]

    def get_id_by_token(self, token):
        if token not in self.token2idx:
            return self.token2idx['<unk>']
        return self.token2idx[token]

    def get_vocab_size(self):
        return len(self.token2idx)

    def __call__(self, report, dataset='iu_xray'):
        if dataset == 'iu_xray':
            tokens = self.clean_report_iu_xray(report).split()
        else:
            tokens = self.clean_report_mimic_cxr(report).split()
        ids = []
        for token in tokens:
            ids.append(self.get_id_by_token(token))
        ids = [0] + ids + [0]
        return ids

    def decode(self, ids):
        txt = ''
        for i, idx in enumerate(ids):
            if idx > 0:
                if i >= 1:
                    txt += ' '
                txt += self.idx2token[idx]
            else:
                break
        return txt

    def decode_batch(self, ids_batch):
        out = []
        for ids in ids_batch:
            out.append(self.decode(ids))
        return out



#!/usr/bin/env python
#
# File Name : bleu.py
#
# Description : Wrapper for BLEU scorer.
#
# Creation Date : 06-01-2015
# Last Modified : Thu 19 Mar 2015 09:13:28 PM PDT
# Authors : Hao Fang <hfang@uw.edu> and Tsung-Yi Lin <tl483@cornell.edu>

# Last modified : Wed 22 May 2019 08:10:00 PM EDT
# By Sabarish Sivanath
# To support Python 3

from .bleu_scorer import BleuScorer


class Bleu:
    def __init__(self, n=4):
        # default compute Blue score up to 4
        self._n = n
        self._hypo_for_image = {}
        self.ref_for_image = {}

    def compute_score(self, gts, res, score_option = 'closest', verbose = 1):
        '''
        Inputs:
            gts - ground truths
            res - predictions
            score_option - {shortest, closest, average}
            verbose - 1 or 0
        Outputs:
            Blue scores
        '''
        assert(gts.keys() == res.keys())
        imgIds = gts.keys()

        bleu_scorer = BleuScorer(n=self._n)
        for id in imgIds:
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert(type(hypo) is list)
            assert(len(hypo) == 1)
            assert(type(ref) is list)
            #assert(len(ref) >= 1)

            bleu_scorer += (hypo[0], ref)

        score, scores = bleu_scorer.compute_score(option = score_option, verbose =verbose)

        # return (bleu, bleu_info)
        return score, scores

    def method(self):
        return "Bleu"



"""
vision language model for MRG task.
It is a baseline model based on transformer (L3H8N512).
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from . import misc
from modules.backbones.transformer import TextEmbed, Decoder, Encoder
from modules.backbones.vision_transformer import VisEmbed
from .beam_search import BeamSearch


def get_hv_mask(hv):
    """
    bidirectional-attention mask for image
    """
    v_masks = hv.new_ones(hv.shape[:2], dtype=torch.long)
    v_masks = v_masks.unsqueeze(-2)
    return v_masks


def get_ht_mask(seq=None):
    """
    unidirectional-attention mask for text
    """
    if seq is not None:
        # crop the last one
        seq = seq[:, :-1]
        seq_mask = (seq.data > 0)
        seq_mask[:, 0] += True
        seq_mask = seq_mask.unsqueeze(-2)
        seq_mask = seq_mask & misc.subsequent_mask(seq.size(-1)).to(seq_mask)
    else:
        seq_mask = None
    return seq_mask, seq


class Baseline(nn.Module):
    def __init__(self, cfgs, tokenizer):
        super(Baseline, self).__init__()
        self.cfgs = cfgs
        self.vocab_size = len(tokenizer.idx2token)
        self.tokenizer = tokenizer

        self.en_num_layers = cfgs["en_num_layers"]
        self.de_num_layers = cfgs["de_num_layers"]
        self.embed_dim = cfgs["embed_dim"]
        self.num_heads = cfgs["num_heads"]
        self.ff_dim = cfgs["embed_dim"] * 4
        self.dropout = cfgs["dropout"]

        # ------------------------------------------
        # Embedding
        self.vis_embed = VisEmbed(embed_dim=self.embed_dim, dropout=self.dropout)
        self.text_embed = TextEmbed(embed_dim=self.embed_dim, vocab_size=self.vocab_size + 1, dropout=self.dropout)

        # Encoder
        self.encoder = Encoder(embed_dim=self.embed_dim, num_layer=self.en_num_layers, num_heads=self.num_heads,
                               ff_dim=self.ff_dim, dropout=self.dropout)
        self.decoder = Decoder(embed_dim=self.embed_dim, num_layer=self.de_num_layers,
                               num_heads=self.num_heads, ff_dim=self.ff_dim, dropout=self.dropout)

        self.logit = nn.Linear(self.embed_dim, self.vocab_size + 1)

        self.cls_token = nn.Parameter(torch.zeros(1, 1, self.embed_dim))
        torch.nn.init.normal_(self.cls_token, std=.02)

        self.beam_search = BeamSearch(cfgs, self.vocab_size)

        # ------------------------------------------
        if cfgs["dataset_name"] == 'iu_xray':
            self.forward = self.forward_iu_xray
        elif cfgs["dataset_name"] == 'mimic_cxr':
            self.forward = self.forward_mimic_cxr
        elif cfgs["dataset_name"] == 'ffa_ir':
            self.forward = self.forward_ffa_ir
        else:
            raise ValueError

        # initialize nn.Linear and nn.LayerNorm
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward_iu_xray(self, images, targets=None, mode='train'):
        B = images.size(0)
        hv = self.vis_embed(images.reshape(B*2, 3, 224, 224))
        outputs = self._forward(hv, targets, mode, B)
        return outputs

    def forward_ffa_ir(self, images, targets=None, mode='train'):
        # B, N, C, H, W
        B, N, C, H, W = images.size()
        images = images.reshape([-1, C, H, W])
        hv = self.vis_embed(images)
        outputs = self._forward(hv, targets, mode, B)
        return outputs

    def forward_mimic_cxr(self, images, targets=None, mode='train'):
        hv = self.vis_embed(images)
        outputs = self._forward(hv, targets, mode, images.size(0))
        return outputs

    def _forward(self, hv, targets, mode, B):
        # append cls token
        hv = hv.reshape([B, -1, self.embed_dim])
        cls_token = self.cls_token + self.vis_embed.pos_embed[:, :1, :]
        cls_tokens = cls_token.expand(hv.shape[0], -1, -1)
        hv = torch.cat((cls_tokens, hv), dim=1)
        # encode
        hv_mask = get_hv_mask(hv)
        hv = self.encoder(hv, hv_mask)
        # decode
        if mode == 'train':
            ht_mask, targets = get_ht_mask(targets)
            ht = self.text_embed(targets)  # [B, L] -> [B, L, D]
            out = self.decoder(ht, hv, self_mask=ht_mask, cross_mask=hv_mask)
            outputs = [F.log_softmax(self.logit(out), dim=-1)]
        elif mode == 'sample':
            self.beam_search.load_model(self.sample_forward,
                                        self.logit)
            outputs, _ = self.beam_search.sample_beam(hv)
            self.beam_search.clean_model()
        else:
            raise ValueError
        return outputs

    def sample_forward(self, hv, ht, v_mask, t_mask):
        ht = self.text_embed(ht)
        return self.decoder(ht, hv, self_mask=t_mask, cross_mask=v_mask)



"""
pipeline for medical report generation
"""
import pandas as pd
from .base import BasePipeline
import numpy as np
import torch
import time
import os


class MRGPipeline(BasePipeline):
    def __init__(self, model, cfgs, 
            criterion=None, 
            metric_caculator=None, 
            optimizer=None, 
            lr_scheduler=None, 
            train_dataloader=None, 
            val_dataloader=None, 
            test_dataloader=None
    ):
        super(MRGPipeline, self).__init__(model, cfgs)

        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.test_dataloader = test_dataloader

        self.metric_caculator = metric_caculator
        self.optimizer = optimizer
        self.criterion = criterion
        self.lr_scheduler = lr_scheduler

    def _train_epoch(self, epoch):
        self.model.train()
        start_time = time.time()
        for batch_idx, (images_id, images, reports_ids, reports_masks) in enumerate(self.train_dataloader):
            images, reports_ids, reports_masks = images.cuda(), reports_ids.cuda(), reports_masks.cuda()
            output = self.model(images, reports_ids, mode='train')
            loss = self.criterion(output, reports_ids, reports_masks)
            self.optimizer.zero_grad()
            loss.backward()
            self.monitor.log_mean(('train loss', loss.item()))
            torch.nn.utils.clip_grad_value_(self.model.parameters(), 0.1)
            self.optimizer.step()

            self.monitor.print(f"\repoch: {epoch} {batch_idx+1}/{len(self.train_dataloader)}\tloss: {loss:.3f}\tmean loss: {self.monitor.name2val['train loss']:.3f}", flush=True)

            if self.cfgs["lr_scheduler"] != 'StepLR':
                self.lr_scheduler.step()
        if self.cfgs["lr_scheduler"] == 'StepLR':
            self.lr_scheduler.step()

        self.monitor.print(f"\nEpoch {epoch}\ttime: {time.time() - start_time:.1f}s")

        self.model.eval()
        with torch.no_grad():
            val_gts, val_res = [], []
            temp_report_ids = []
            for batch_idx, (images_id, images, reports_ids, reports_masks) in enumerate(self.val_dataloader):
                images, reports_ids, reports_masks = images.cuda(), reports_ids.cuda(), reports_masks.cuda()
                output = self.model(images, mode='sample')
                reports = self.model.tokenizer.decode_batch(output.cpu().numpy())
                ground_truths = self.model.tokenizer.decode_batch(reports_ids[:, 1:].cpu().numpy())
                val_res.extend(reports)
                val_gts.extend(ground_truths)
                temp_report_ids.extend(output.cpu().numpy())
                self.monitor.print(f"\rVal Processing: [{int((batch_idx + 1) / len(self.val_dataloader) * 100)}%]", flush=True)
            _, n_reports = count_report_pattern(temp_report_ids)
            val_met = self.metric_caculator.compute_scores({i: [gt] for i, gt in enumerate(val_gts)},
                                                           {i: [re] for i, re in enumerate(val_res)})
            # record val metrics
            val_met['n_reports'] = n_reports
            for k, v in val_met.items():
                self.monitor.log(('val_' + k, v))

        self.model.eval()
        with torch.no_grad():
            test_gts, test_res, p = [], [], []
            temp_report_ids = []
            for batch_idx, (images_id, images, reports_ids, reports_masks) in enumerate(self.test_dataloader):
                images, reports_ids, reports_masks = images.cuda(), reports_ids.cuda(), reports_masks.cuda()
                output = self.model(images, mode='sample')
                reports = self.model.tokenizer.decode_batch(output.cpu().numpy())
                ground_truths = self.model.tokenizer.decode_batch(reports_ids[:, 1:].cpu().numpy())
                test_res.extend(reports)
                test_gts.extend(ground_truths)
                temp_report_ids.extend(output.cpu().numpy())
                self.monitor.print(f"\rTest Processing: [{int((batch_idx + 1) / len(self.test_dataloader) * 100)}%]", flush=True)
            _, n_reports = count_report_pattern(temp_report_ids)
            test_met = self.metric_caculator.compute_scores({i: [gt] for i, gt in enumerate(test_gts)},
                                                            {i: [re] for i, re in enumerate(test_res)})

            test_met['n_reports'] = n_reports
            for k, v in test_met.items():
                self.monitor.log(('test_' + k, v))

        if self.cfgs['monitor_metric_curves']:
            self.monitor.plot_current_metrics(epoch, self.monitor.name2val)


    def inference(self, loader):
        """
        This function takes an integer and a string as input and returns a boolean.
        
        Args:
            loader: 
        
        Returns:
            dict: the score of each metric
        """
        self.model.eval()
        with torch.no_grad():
            test_gts, test_res, report_pattern, img_ids = [], [], [], []
            for batch_idx, (images_id, images, reports_ids, reports_masks) in enumerate(loader):
                images, reports_ids, reports_masks = images.cuda(), reports_ids.cuda(), reports_masks.cuda()
                output = self.model(images, mode='sample')
                reports = self.model.tokenizer.decode_batch(output.cpu().numpy())

                ground_truths = self.model.tokenizer.decode_batch(reports_ids[:, 1:].cpu().numpy())
                test_res.extend(reports)
                test_gts.extend(ground_truths)
                self.monitor.print(f"\rInference Processing: [{int((batch_idx + 1) / len(loader) * 100)}%]", flush=True)
            # test_res = torch.load("results/mimic_cxr/DMIRG/DMIRG/118_report_100.npy")
            test_met = self.metric_caculator.compute_scores({i: [gt] for i, gt in enumerate(test_gts)},
                                                            {i: [re] for i, re in enumerate(test_res)})
            
            # save_report(test_res, test_gts, img_ids, os.path.join(self.checkpoint_dir, 'report.csv'))
            self.monitor.print(f"\n{test_met}")
            return test_met

def count_report_pattern(report_pattern):
    report_pattern = np.array(report_pattern)
    unique_report_pattern = np.unique(report_pattern, axis=0)
    num_report_pattern = unique_report_pattern.shape[0]
    return unique_report_pattern, num_report_pattern




from hcpcvlr.utils.cfgs_loader import load_yaml
from hcpcvlr.api.pipeline import ChatPipeline
from hcpcvlr.utils.metrics import MetricCalculator
from hcpcvlr.data import load_scienceqa_data
from hcpcvlr.models.chat.caco_cot import CaCoCoT


cfgs = load_yaml("configs/chat/CaCo_CoT.yaml")
# print(cfgs)
model = CaCoCoT(cfgs)
# -------------------
# inference on Science QA
# -------------------
work = ChatPipeline(model, cfgs, metric_caculator=MetricCalculator(cfgs))
work.inference()



"""
Visual Causal Intervention for Radiology Report Generation
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

from modules.backbones.transformer import Encoder, DecoderLayer, LayerNorm
from .baseline import Baseline, get_ht_mask, get_hv_mask
from . import misc
from modules.layers.vlci import LocalSample, GlobalSample, VDM, LDM


def get_cross_mask(hv, seq):
    seq = seq[:, :-1]
    seq_mask = (seq.data > 0)
    seq_mask[:, 0] += True
    seq_mask = seq_mask.unsqueeze(-1)
    v_mask = hv.new_ones(hv.shape[:2], dtype=torch.long)
    v_mask = v_mask.unsqueeze(-2)

    cross_mask = seq_mask & v_mask
    return cross_mask


class CausalEncoder(Encoder):
    def __init__(self, embed_dim, num_layer, num_heads, ff_dim, dropout, mode='n'):
        super(CausalEncoder, self).__init__(embed_dim, num_layer, num_heads, ff_dim, dropout)
        self.mode = mode
        self.do = VDM(embed_dim, num_heads, ff_dim, dropout)

        self.global_sample = GlobalSample(embed_dim, num_heads, ff_dim, dropout)
        self.local_sample = LocalSample(embed_dim, num_heads, ff_dim, dropout)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, h, mask=None, pos=None, k=6, proj=False):
        attn = []
        for layer in self.layers:
            h = layer(h, mask)
            attn.append(layer.attn.attn)

        if h.size(1) > 197:
            h = h + torch.cat([pos, pos[:, 1:, :]], dim=1)
        else:
            h = h + pos

        fl = self.local_sample(h, attn, k)
        fg = self.global_sample(h)
        mediator = {"local": fl, "global": fg, 'attn': attn}
        h = self.do(h, fl=fl, fg=fg, mode=self.mode, proj=proj)

        h = self.norm(h)
        return h, mediator


class CausalDecoder(nn.Module):
    def __init__(self, embed_dim, num_layer, num_heads, ff_dim, dropout, mode='n'):
        super(CausalDecoder, self).__init__()
        self.mode = mode
        self.norm = LayerNorm(embed_dim)
        self.do = LDM(embed_dim)
        self.layers = nn.ModuleList(
            [DecoderLayer(embed_dim, num_heads, ff_dim, dropout) for _ in range(num_layer)])
        self.ht = None

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, output, h, self_mask=None, cross_mask=None, z=None, fl=None, proj=False):

        output = self.do(output, z, fl, self.mode, proj)
        self.ht = output
        for i in range(len(self.layers)):
            output = self.layers[i](output, h, self_mask, cross_mask)
            # attn.append(self.layers[i].cross_attn.attn)

        output = self.norm(output)
        return output


class VLCI(Baseline):
    def __init__(self, cfgs, tokenizer):
        super(VLCI, self).__init__(cfgs, tokenizer)
        # Encoder
        self.z = None
        self.fl = None
        self.vocab = torch.arange(0, self.vocab_size + 1).unsqueeze(-1).long().cuda()
        self.z_norm = LayerNorm(self.embed_dim)
        self.fl_norm = LayerNorm(self.embed_dim)

        self.encoder = CausalEncoder(embed_dim=self.embed_dim, num_layer=self.en_num_layers, num_heads=self.num_heads,
                                     ff_dim=self.ff_dim, dropout=self.dropout, mode=cfgs["v_causal"])

        self.decoder = CausalDecoder(embed_dim=self.embed_dim, num_layer=self.de_num_layers, num_heads=self.num_heads,
                                     ff_dim=self.ff_dim, dropout=self.dropout, mode=cfgs["l_causal"])

    def _forward(self, hv, targets, mode, B):
        # append cls token
        hv = hv.reshape([B, -1, self.embed_dim])
        cls_token = self.cls_token + self.vis_embed.pos_embed[:, :1, :]
        cls_tokens = cls_token.expand(hv.shape[0], -1, -1)
        hv = torch.cat((cls_tokens, hv), dim=1)
        # encode
        hv_mask = get_hv_mask(hv)
        hv, mediator = self.encoder(hv, hv_mask, self.vis_embed.pos_embed, k=self.cfgs['local_sample'], proj=self.cfgs['v_causal_proj'])

        if self.cfgs["l_causal"]:
            z = self.text_embed.word_embed(self.vocab).reshape(1, -1, hv.size(-1)).detach()
            z = self.z_norm(z)
            self.fl = self.fl_norm(mediator['local'])
            self.z = z.repeat(hv.size(0), 1, 1)

        # decode
        if mode == 'train':
            cross_mask = get_cross_mask(hv, targets)
            ht_mask, targets = get_ht_mask(targets)
            ht = self.text_embed(targets)  # [B, L] -> [B, L, D]
            out = self.decoder(ht, hv, self_mask=ht_mask, cross_mask=cross_mask,
                               z=self.z, fl=self.fl, proj=self.cfgs['l_causal_proj'])
            outputs = F.log_softmax(self.logit(out), dim=-1)
            outputs = [outputs, mediator, {'ht': self.decoder.ht.detach(), 'hv': hv.detach()}]
        elif mode == 'sample':
            self.beam_search.load_model(self.sample_forward, self.logit)
            outputs, _ = self.beam_search.sample_beam(hv)
            self.beam_search.clean_model()
        else:
            raise ValueError
        return outputs

    def sample_forward(self, hv, ht, v_mask, t_mask):
        ht = self.text_embed(ht)
        v_mask = get_hv_mask(hv)
        if self.z is not None and self.z.size(0) != ht.size(0):
            self.z = misc.repeat_tensors(self.cfgs["beam_size"], self.z)
            self.fl = misc.repeat_tensors(self.cfgs["beam_size"], self.fl)
        out = self.decoder(ht, hv, self_mask=t_mask, cross_mask=v_mask,
                           z=self.z, fl=self.fl, proj=self.cfgs['l_causal_proj'])
        return out



"""
modules of transformer
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from ..utils.pos_embed import get_1d_sincos_pos_embed


def clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


def attention(query, key, value, mask=None, dropout=None):
    d_k = query.size(-1)
    scores = torch.matmul(query, key.transpose(-2, -1)) / math.sqrt(d_k)
    if mask is not None:
        scores = scores.masked_fill(mask == 0, -1e9)
    p_attn = F.softmax(scores, dim=-1)
    if dropout is not None:
        p_attn = dropout(p_attn)
    return torch.matmul(p_attn, value), p_attn


def subsequent_mask(size):
    attn_shape = (1, size, size)
    subsequent_mask = np.triu(np.ones(attn_shape), k=1).astype('uint8')
    return torch.from_numpy(subsequent_mask) == 0


class LayerNorm(nn.Module):
    def __init__(self, features, eps=1e-6):
        super(LayerNorm, self).__init__()
        self.gamma = nn.Parameter(torch.ones(features))
        self.beta = nn.Parameter(torch.zeros(features))
        self.eps = eps

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.gamma * (x - mean) / (std + self.eps) + self.beta


class MultiHeadedAttention(nn.Module):
    def __init__(self, h, d_model, dropout=0.1):
        super(MultiHeadedAttention, self).__init__()
        assert d_model % h == 0
        self.d_k = d_model // h
        self.h = h
        self.linears = clones(nn.Linear(d_model, d_model), 4)
        self.attn = None
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, query, key, value, mask=None):
        if mask is not None:
            mask = mask.unsqueeze(1)
        nbatches = query.size(0)
        query, key, value = \
            [l(x).view(nbatches, -1, self.h, self.d_k).transpose(1, 2)
             for l, x in zip(self.linears, (query, key, value))]

        x, self.attn = attention(query, key, value, mask=mask, dropout=self.dropout)

        x = x.transpose(1, 2).contiguous().view(nbatches, -1, self.h * self.d_k)
        return self.linears[-1](x)


class PositionwiseFeedForward(nn.Module):
    def __init__(self, d_model, d_ff, dropout=0.1):
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.w_2(self.dropout(F.relu(self.w_1(x))))


class WordEmbed(nn.Module):
    def __init__(self, vocab_size, embed_dim):
        super(WordEmbed, self).__init__()
        self.lut = nn.Embedding(vocab_size, embed_dim)
        self.embed_dim = embed_dim

    def forward(self, x):
        return self.lut(x) * math.sqrt(self.embed_dim)


class TextEmbed(nn.Module):
    """
    test embedding with 1d sin-cos embedding
    """

    def __init__(self, embed_dim, vocab_size, dropout=0.):
        super(TextEmbed, self).__init__()
        self.word_embed = WordEmbed(vocab_size, embed_dim)
        self.pos_embed = nn.Parameter(torch.zeros(1, 5000 + 1, embed_dim),
                                      requires_grad=False)  # fixed sin-cos embedding
        self.dropout = nn.Dropout(p=dropout)
        self.norm = LayerNorm(embed_dim)
        
        self.initialize_weights()
        # initialize nn.Linear and nn.LayerNorm
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
    
    def initialize_weights(self):
        # initialization
        pos_embed = get_1d_sincos_pos_embed(self.pos_embed.shape[-1], cls_token=True)
        self.pos_embed.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))

    def forward(self, x):
        x = self.word_embed(x)
        x = x + self.pos_embed[:, 1:x.size(1)+1]
        x = self.dropout(self.norm(x))
        return x


class SublayerConnection(nn.Module):
    def __init__(self, embed_dim, dropout):
        super(SublayerConnection, self).__init__()
        self.norm = LayerNorm(embed_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, sublayer):
        return x + self.dropout(sublayer(self.norm(x)))


class EncoderLayer(nn.Module):
    def __init__(self, embed_dim, num_heads, ff_dim, dropout):
        super(EncoderLayer, self).__init__()
        self.attn = MultiHeadedAttention(num_heads, embed_dim, dropout)
        self.feed_forward = PositionwiseFeedForward(embed_dim, ff_dim, dropout)
        self.sublayer_attn = SublayerConnection(embed_dim, dropout)
        self.sublayer_ff = SublayerConnection(embed_dim, dropout)

    def forward(self, x, mask=None):
        x = self.sublayer_attn(x, lambda x: self.attn(x, x, x, mask))
        x = self.sublayer_ff(x, self.feed_forward)
        return x


class Encoder(nn.Module):
    def __init__(self, embed_dim, num_layer, num_heads, ff_dim, dropout):
        super(Encoder, self).__init__()
        self.layers = nn.ModuleList([EncoderLayer(embed_dim, num_heads, ff_dim, dropout)
                                     for _ in range(num_layer)])
        self.norm = LayerNorm(embed_dim)

    def forward(self, h, mask=None):
        for layer in self.layers:
            h = layer(h, mask)
        h = self.norm(h)
        return h


class DecoderLayer(nn.Module):
    def __init__(self, embed_dim, num_heads, ff_dim, dropout):
        super(DecoderLayer, self).__init__()
        self.self_attn = MultiHeadedAttention(num_heads, embed_dim, dropout)
        self.cross_attn = MultiHeadedAttention(num_heads, embed_dim, dropout)
        self.feed_forward = PositionwiseFeedForward(embed_dim, ff_dim, dropout)
        self.sublayer_cross = SublayerConnection(embed_dim, dropout)
        self.sublayer_self = SublayerConnection(embed_dim, dropout)
        self.sublayer_ff = SublayerConnection(embed_dim, dropout)

    def forward(self, x, h, self_mask=None, cross_mask=None):
        x = self.sublayer_self(x, lambda x: self.self_attn(x, x, x, self_mask))
        x = self.sublayer_cross(x, lambda x: self.cross_attn(x, h, h, cross_mask))
        x = self.sublayer_ff(x, self.feed_forward)
        return x


class Decoder(nn.Module):
    def __init__(self, embed_dim, num_layer, num_heads, ff_dim, dropout):
        super(Decoder, self).__init__()
        self.layers = nn.ModuleList([DecoderLayer(embed_dim, num_heads, ff_dim, dropout)
                                     for _ in range(num_layer)])
        self.norm = LayerNorm(embed_dim)

    def forward(self, x, h, self_mask=None, cross_mask=None):
        for i in range(len(self.layers)):
            x = self.layers[i](x, h, self_mask, cross_mask)
        x = self.norm(x)
        return x


class Transformer(nn.Module):

    arch_settings = {
        "nano": (3, 8, 512),
        "small": (6, 8, 512),
        "base": (12, 12, 768),
        "large": (24, 16, 1024),
        "huge": (32, 16, 1280)
    }

    def __init__(self):
        super().__init__()

    def forward(self):
        pass



"""
Language Model Generation loss function
"""
import torch
import torch.nn as nn
from torch.autograd import Variable
import torch.nn.functional as F


class LanguageModelCriterion(nn.Module):
    def __init__(self):
        super(LanguageModelCriterion, self).__init__()

    def forward(self, input, target, mask):
        # truncate to the same size
        target = target[:, :input.size(1)]
        mask = mask[:, :input.size(1)]
        output = -input.gather(2, target.long().unsqueeze(2)).squeeze(2) * mask
        output = torch.sum(output) / torch.sum(mask)

        return output


def compute_lm_loss(output, reports_ids, reports_masks):
    if isinstance(output, list):
        output = output[0]
    criterion = LanguageModelCriterion()
    loss = criterion(output, reports_ids[:, 1:], reports_masks[:, 1:]).mean()
    return loss



"""
beam search for language model sample mode
"""
import torch
import torch.nn.functional as F
from . import misc


class BeamSearch(object):
    def __init__(self, cfgs, vocab_size):
        super(BeamSearch, self).__init__()
        self.cfgs = cfgs
        self.beam_size = cfgs["beam_size"]  # k
        self.group_size = cfgs["group_size"]
        self.sample_n = cfgs["sample_n"]
        self.vocab_size = vocab_size
        self.done_beams = None
        self.model = None
        self.logit = None

    def load_model(self, model, logit):
        self.model = model
        self.logit = logit

    def clean_model(self):
        self.model = None
        self.logit = None

    def get_logprobs_state(self, it, att_feats, att_masks, state, output_logsoftmax=1):
        # 'it' contains a word index

        if len(state) == 0:
            ys = it.unsqueeze(1)
        else:
            ys = torch.cat([state[0][0], it.unsqueeze(1)], dim=1)

        out = self.model(att_feats, ys, att_masks, misc.subsequent_mask(ys.size(1)).cuda())

        output = out[:, -1]
        state = [ys.unsqueeze(0)]
        if output_logsoftmax:
            logprobs = F.log_softmax(self.logit(output), dim=1)
        else:
            logprobs = self.logit(output)

        return logprobs, state

    @staticmethod
    def _prepare_feature(att_feats, att_masks=None, seq=None):

        if att_masks is None:
            att_masks = att_feats.new_ones(att_feats.shape[:2], dtype=torch.long)
        att_masks = att_masks.unsqueeze(-2)

        if seq is not None:
            # crop the last one
            seq = seq[:, :-1]
            seq_mask = (seq.data > 0)
            seq_mask[:, 0] += True
            seq_mask = seq_mask.unsqueeze(-2)
            seq_mask = seq_mask & misc.subsequent_mask(seq.size(-1)).to(seq_mask)
        else:
            seq_mask = None

        return att_masks, seq_mask

    def sample_beam(self, att_feats, att_masks=None):
        # when sample_n == beam_size then each beam is a sample.
        assert self.sample_n == 1 or self.sample_n == self.beam_size // self.group_size, \
            'when beam search, sample_n == 1 or beam search'
        batch_size = att_feats.size(0)

        att_masks, _ = self._prepare_feature(att_feats, att_masks)

        assert self.beam_size <= self.vocab_size + 1, 'lets assume this for now, otherwise this corner case causes a few headaches down the road. can be dealt with in future if needed'
        seq = att_feats.new_full((batch_size * self.sample_n, self.cfgs["max_seq_length"]), self.cfgs["pad_idx"], dtype=torch.long)
        seqLogprobs = att_feats.new_zeros(batch_size * self.sample_n, self.cfgs["max_seq_length"], self.vocab_size + 1)
        # lets process every image independently for now, for simplicity

        self.done_beams = [[] for _ in range(batch_size)]

        state = []

        # first step, feed bos
        it = att_feats.new_full([batch_size], self.cfgs["bos_idx"], dtype=torch.long)
        logprobs, state = self.get_logprobs_state(it, att_feats, att_masks, state)

        att_feats, att_masks = misc.repeat_tensors(self.beam_size, [att_feats, att_masks])
        self.done_beams = self.beam_search(state, logprobs, att_feats, att_masks)
        for k in range(batch_size):
            if self.sample_n == self.beam_size:
                for _n in range(self.sample_n):
                    seq_len = self.done_beams[k][_n]['seq'].shape[0]
                    seq[k * self.sample_n + _n, :seq_len] = self.done_beams[k][_n]['seq']
                    seqLogprobs[k * self.sample_n + _n, :seq_len] = self.done_beams[k][_n]['logps']
            else:
                seq_len = self.done_beams[k][0]['seq'].shape[0]
                seq[k, :seq_len] = self.done_beams[k][0]['seq']  # the first beam has highest cumulative score
                seqLogprobs[k, :seq_len] = self.done_beams[k][0]['logps']
        # return the samples and their log likelihoods
        return seq, seqLogprobs

    def beam_search(self, init_state, init_logprobs, *args):

        # function computes the similarity score to be augmented
        def add_diversity(beam_seq_table, logprobs, t, divm, diversity_lambda, bdash):
            local_time = t - divm
            unaug_logprobs = logprobs.clone()
            batch_size = beam_seq_table[0].shape[0]

            if divm > 0:
                change = logprobs.new_zeros(batch_size, logprobs.shape[-1])
                for prev_choice in range(divm):
                    prev_decisions = beam_seq_table[prev_choice][:, :, local_time]  # Nxb
                    for prev_labels in range(bdash):
                        change.scatter_add_(1, prev_decisions[:, prev_labels].unsqueeze(-1),
                                            change.new_ones(batch_size, 1))

                if local_time == 0:
                    logprobs = logprobs - change * diversity_lambda
                else:
                    logprobs = logprobs - self.repeat_tensor(bdash, change) * diversity_lambda

            return logprobs, unaug_logprobs

        # does one step of classical beam search

        def beam_step(logprobs, unaug_logprobs, beam_size, t, beam_seq, beam_seq_logprobs, beam_logprobs_sum, state):
            # INPUTS:
            # logprobs: probabilities augmented after diversity N*bxV
            # beam_size: obvious
            # t        : time instant
            # beam_seq : tensor contanining the beams
            # beam_seq_logprobs: tensor contanining the beam logprobs
            # beam_logprobs_sum: tensor contanining joint logprobs
            # OUPUTS:
            # beam_seq : tensor containing the word indices of the decoded captions Nxbxl
            # beam_seq_logprobs : log-probability of each decision made, NxbxlxV
            # beam_logprobs_sum : joint log-probability of each beam Nxb

            batch_size = beam_logprobs_sum.shape[0]
            vocab_size = logprobs.shape[-1]
            logprobs = logprobs.reshape(batch_size, -1, vocab_size)  # NxbxV
            if t == 0:
                assert logprobs.shape[1] == 1
                beam_logprobs_sum = beam_logprobs_sum[:, :1]
            candidate_logprobs = beam_logprobs_sum.unsqueeze(-1) + logprobs  # beam_logprobs_sum Nxb logprobs is NxbxV
            ys, ix = torch.sort(candidate_logprobs.reshape(candidate_logprobs.shape[0], -1), -1, True)
            ys, ix = ys[:, :beam_size], ix[:, :beam_size]
            beam_ix = ix // vocab_size  # Nxb which beam
            selected_ix = ix % vocab_size  # Nxb # which world
            state_ix = (beam_ix + torch.arange(batch_size).type_as(beam_ix).unsqueeze(-1) * logprobs.shape[1]).reshape(
                -1)  # N*b which in Nxb beams

            if t > 0:
                # gather according to beam_ix
                assert (beam_seq.gather(1, beam_ix.unsqueeze(-1).expand_as(beam_seq)) ==
                        beam_seq.reshape(-1, beam_seq.shape[-1])[state_ix].view_as(beam_seq)).all()
                beam_seq = beam_seq.gather(1, beam_ix.unsqueeze(-1).expand_as(beam_seq))

                beam_seq_logprobs = beam_seq_logprobs.gather(1, beam_ix.unsqueeze(-1).unsqueeze(-1).expand_as(
                    beam_seq_logprobs))

            beam_seq = torch.cat([beam_seq, selected_ix.unsqueeze(-1)], -1)  # beam_seq Nxbxl
            beam_logprobs_sum = beam_logprobs_sum.gather(1, beam_ix) + \
                                logprobs.reshape(batch_size, -1).gather(1, ix)
            assert (beam_logprobs_sum == ys).all()
            _tmp_beam_logprobs = unaug_logprobs[state_ix].reshape(batch_size, -1, vocab_size)
            beam_logprobs = unaug_logprobs.reshape(batch_size, -1, vocab_size).gather(1,
                                                                                      beam_ix.unsqueeze(-1).expand(-1,
                                                                                                                   -1,
                                                                                                                   vocab_size))  # NxbxV
            assert (_tmp_beam_logprobs == beam_logprobs).all()
            beam_seq_logprobs = torch.cat([
                beam_seq_logprobs,
                beam_logprobs.reshape(batch_size, -1, 1, vocab_size)], 2)

            new_state = [None for _ in state]
            for _ix in range(len(new_state)):
                #  copy over state in previous beam q to new beam at vix
                new_state[_ix] = state[_ix][:, state_ix]
            state = new_state
            return beam_seq, beam_seq_logprobs, beam_logprobs_sum, state

        # Start diverse_beam_search
        opt = self.cfgs
        temperature = opt["temperature"]  # This should not affect beam search, but will affect dbs
        beam_size = opt["beam_size"]
        group_size = opt["group_size"]
        diversity_lambda = 0.5 if opt["diversity_lambda"] is None else opt["diversity_lambda"]
        decoding_constraint = opt["decoding_constraint"]
        suppress_UNK = 0 if opt["suppress_UNK"] is None else opt["suppress_UNK"]
        length_penalty = misc.penalty_builder('' if opt["length_penalty"] is None else opt["length_penalty"])
        bdash = beam_size // group_size  # beam per group

        batch_size = init_logprobs.shape[0]
        device = init_logprobs.device
        # INITIALIZATIONS
        beam_seq_table = [torch.LongTensor(batch_size, bdash, 0).to(device) for _ in range(group_size)]
        beam_seq_logprobs_table = [torch.FloatTensor(batch_size, bdash, 0, self.vocab_size + 1).to(device) for _ in
                                   range(group_size)]
        beam_logprobs_sum_table = [torch.zeros(batch_size, bdash).to(device) for _ in range(group_size)]

        # logprobs # logprobs predicted in last time step, shape (beam_size, vocab_size+1)
        done_beams_table = [[[] for __ in range(group_size)] for _ in range(batch_size)]
        state_table = [[_.clone() for _ in init_state] for _ in range(group_size)]
        logprobs_table = [init_logprobs.clone() for _ in range(group_size)]
        # END INIT

        # Chunk elements in the args
        args = list(args)
        args = misc.split_tensors(group_size, args)  # For each arg, turn (Bbg)x... to (Bb)x(g)x...
        if self.__class__.__name__ == 'AttEnsemble':
            args = [[[args[j][i][k] for i in range(len(self.models))] for j in range(len(args))] for k in
                    range(group_size)]  # group_name, arg_name, model_name
        else:
            args = [[args[i][j] for i in range(len(args))] for j in range(group_size)]

        for t in range(self.cfgs["max_seq_length"] + group_size - 1):
            for divm in range(group_size):
                if t >= divm and t <= self.cfgs["max_seq_length"] + divm - 1:
                    # add diversity
                    logprobs = logprobs_table[divm]
                    # suppress previous word
                    if decoding_constraint and t - divm > 0:
                        logprobs.scatter_(1, beam_seq_table[divm][:, :, t - divm - 1].reshape(-1, 1).to(device),
                                          float('-inf'))
                    # suppress UNK tokens in the decoding
                    if suppress_UNK and hasattr(self, 'vocab') and self.vocab[str(logprobs.size(1) - 1)] == 'UNK':
                        logprobs[:, logprobs.size(1) - 1] = logprobs[:, logprobs.size(1) - 1] - 1000
                        # diversity is added here
                    # the function directly modifies the logprobs values and hence, we need to return
                    # the unaugmented ones for sorting the candidates in the end. # for historical
                    # reasons :-)
                    logprobs, unaug_logprobs = add_diversity(beam_seq_table, logprobs, t, divm, diversity_lambda, bdash)

                    # infer new beams
                    beam_seq_table[divm], \
                    beam_seq_logprobs_table[divm], \
                    beam_logprobs_sum_table[divm], \
                    state_table[divm] = beam_step(logprobs,
                                                  unaug_logprobs,
                                                  bdash,
                                                  t - divm,
                                                  beam_seq_table[divm],
                                                  beam_seq_logprobs_table[divm],
                                                  beam_logprobs_sum_table[divm],
                                                  state_table[divm])

                    # if time's up... or if end token is reached then copy beams
                    for b in range(batch_size):
                        is_end = beam_seq_table[divm][b, :, t - divm] == self.cfgs["eos_idx"]
                        assert beam_seq_table[divm].shape[-1] == t - divm + 1
                        if t == self.cfgs["max_seq_length"] + divm - 1:
                            is_end.fill_(1)
                        for vix in range(bdash):
                            if is_end[vix]:
                                final_beam = {
                                    'seq': beam_seq_table[divm][b, vix].clone(),
                                    'logps': beam_seq_logprobs_table[divm][b, vix].clone(),
                                    'unaug_p': beam_seq_logprobs_table[divm][b, vix].sum().item(),
                                    'p': beam_logprobs_sum_table[divm][b, vix].item()
                                }
                                final_beam['p'] = length_penalty(t - divm + 1, final_beam['p'])
                                done_beams_table[b][divm].append(final_beam)
                        beam_logprobs_sum_table[divm][b, is_end] -= 1000

                    # move the current group one step forward in time

                    it = beam_seq_table[divm][:, :, t - divm].reshape(-1)
                    logprobs_table[divm], state_table[divm] = self.get_logprobs_state(it.cuda(), *(
                            args[divm] + [state_table[divm]]))
                    logprobs_table[divm] = F.log_softmax(logprobs_table[divm] / temperature, dim=-1)

        # all beams are sorted by their log-probabilities
        done_beams_table = [[sorted(done_beams_table[b][i], key=lambda x: -x['p'])[:bdash] for i in range(group_size)]
                            for b in range(batch_size)]
        done_beams = [sum(_, []) for _ in done_beams_table]
        return done_beams



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
 
    name="hcpcvlr",
    version="0.0.2",
    description="CausalVLR: A Toolbox and Benchmark for Visual-Linguistic Causal Reasoning (视觉-语言因果推理开源框架)",
    url="https://github.com/HCPLab-SYSU/CausalVLR",  
    author=[
            "hcp_lab",
            "liuyang",
            "chenweixing"
    ],
    author_email=[
        "hcp.sysu@gmail.com",
        "liuy856@mail.sysu.edu.cn",
        "chen867820261@gmail.com"
    ], 
    classifiers=[
     
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
   
    keywords="causal",
    # packages=['hcpcvlr'],
    # package_dir={"": "src"},  # Optional 手动指定包目录
    
    packages=find_packages(),
   
    python_requires=">=3.7, <4",  # python 版本要求
 
    install_requires=[
        "torch>=1.6.0",
        "pyyaml",
        "pandas",
        "einops",
        "scipy",
        "matplotlib",
        "dominate",
        "visdom"
        ],
)


"""
pipeline for reasoning via LLM
"""
import pandas as pd
from .base import BasePipeline
import numpy as np
import torch
import random
import time
import os

class ChatPipeline(BasePipeline):
    def __init__(self, model, cfgs, 
            metric_caculator=None, 
            test_dataloader=None
    ):
        random.seed(cfgs["seed"])
        np.random.seed(cfgs["seed"])
        super(ChatPipeline, self).__init__(model, cfgs)
        self.test_dataloader = test_dataloader
        self.metric_caculator = metric_caculator

    def _train_epoch(self, epoch):
        pass

    def inference(self):
        """
        This function takes an integer and a string as input and returns a boolean.
        
        Args:
            loader: 
        
        Returns:
            dict: the score of each metric
        """
        problems, qids, shot_qids = self.test_dataloader(self.cfgs)
        _qid, _answer, _output = [], [], []
        for i in range(len(qids)):
            qid = qids[i]
            choice = problems[qid]["choices"]
            prompt = self.model.build_prompt_sciqa(problems, qid)
            qid, answer, output = self.model(qid, prompt, choice)
            _qid.append(qid)
            _answer.append(answer)
            _output.append(output)
        return _qid, _answer, _output
